"""Combined team model."""

# pylint: disable=too-many-locals,too-many-branches,too-many-statements,too-many-arguments,duplicate-code,too-many-positional-arguments
import functools
from typing import Any

from ..coach_model import CoachModel
from ..news_model import NewsModel
from ..odds_model import OddsModel
from ..player_model import PlayerModel
from ..social_model import SocialModel
from ..team_model import VERSION, TeamModel
from .combined_coach_model import create_combined_coach_model
from .combined_player_model import create_combined_player_model
from .ffill import ffill
from .most_interesting import more_interesting
from .normalise_name import normalise_name


def _compare_player_models(left: PlayerModel, right: PlayerModel) -> int:
    if left.jersey is not None and right.jersey is not None:
        if left.jersey < right.jersey:
            return -1
        if left.jersey > right.jersey:
            return 1
    if left.name < right.name:
        return -1
    if left.name < right.name:
        return 1
    return 0


def create_combined_team_model(
    team_models: list[TeamModel],
    identifier: str,
    player_identity_map: dict[str, str],
    names: dict[str, str],
    coach_names: dict[str, str],
    player_ffill: dict[str, dict[str, Any]],
    team_ffill: dict[str, dict[str, Any]],
    coach_ffill: dict[str, dict[str, Any]],
) -> TeamModel:
    """Create a team model by combining many team models."""
    location = None
    players: dict[str, list[PlayerModel]] = {}
    odds: dict[str, list[OddsModel]] = {}
    news: dict[str, NewsModel] = {}
    social: dict[str, SocialModel] = {}
    coaches: dict[str, list[CoachModel]] = {}
    points = None
    ladder_rank = None
    field_goals = None
    lbw = None
    end_dt = None
    runs = None
    wickets = None
    overs = None
    balls = None
    byes = None
    leg_byes = None
    wides = None
    no_balls = None
    penalties = None
    balls_per_over = None
    fours = None
    sixes = None
    catches = None
    catches_dropped = None
    disposal_efficiency = None
    uncontested_marks = None
    total_possessions = None
    tackles_inside_50 = None
    centre_clearances = None
    stoppage_clearances = None
    goal_accuracy = None
    rushed_behinds = None
    touched_behinds = None
    left_behinds = None
    left_posters = None
    right_behinds = None
    right_posters = None
    total_interchange_count = None
    interchange_count_q1 = None
    interchange_count_q2 = None
    interchange_count_q3 = None
    interchange_count_q4 = None
    game_winning_goals = None
    headed_goals = None
    inaccurate_crosses = None
    inaccurate_long_balls = None
    inaccurate_passes = None
    inaccurate_through_balls = None
    left_footed_shots = None
    longball_percentage = None
    offsides = None
    penalty_kick_goals = None
    penalty_kick_percentage = None
    penalty_kick_shots = None
    penalty_kicks_missed = None
    possession_percentage = None
    possession_time = None
    right_footed_shots = None
    shoot_out_goals = None
    shoot_out_misses = None
    shoot_out_percentage = None
    shot_assists = None
    shot_percentage = None
    shots_headed = None
    shots_off_target = None
    shots_on_post = None
    shots_on_target = None
    through_ball_percentage = None
    total_crosses = None
    total_goals = None
    total_long_balls = None
    total_passes = None
    total_shots = None
    total_through_balls = None
    draws = None
    sub_outs = None
    suspensions = None
    time_ended = None
    time_started = None
    win_percentage = None
    wins = None
    won_corners = None
    yellow_cards = None
    clean_sheet = None
    crosses_caught = None
    goals_conceded = None
    partial_clean_sheet = None
    penalty_kick_conceded = None
    penalty_kick_save_percentage = None
    penalty_kicks_faced = None
    penalty_kicks_saved = None
    punches = None
    saves = None
    shoot_out_kicks_faced = None
    shoot_out_kicks_saved = None
    shoot_out_save_percentage = None
    shots_faced = None
    smothers = None
    unclaimed_crosses = None
    accurate_crosses = None
    accurate_long_balls = None
    accurate_passes = None
    accurate_through_balls = None
    cross_percentage = None
    free_kick_goals = None
    free_kick_percentage = None
    free_kick_shots = None
    game_winning_assists = None
    blocked_shots = None
    effective_clearances = None
    effective_tackles = None
    ineffective_tackles = None
    interceptions = None
    tackle_percentage = None
    appearances = None
    average_rating_from_correspondent = None
    average_rating_from_data_feed = None
    average_rating_from_editor = None
    average_rating_from_user = None
    did_not_play = None
    fouls_committed = None
    fouls_suffered = None
    goal_difference = None
    losses = None
    lost_corners = None
    minutes = None
    own_goals = None
    pass_percentage = None
    red_cards = None
    starts = None
    sub_ins = None
    pitch_count = None
    strikes = None
    strike_pitch_ratio = None
    games_played = None
    team_games_played = None
    double_plays = None
    opportunities = None
    errors = None
    passed_balls = None
    outfield_assists = None
    pickoffs = None
    putouts = None
    outs_on_field = None
    triple_plays = None
    balls_in_zone = None
    extra_bases = None
    outs_made = None
    catcher_third_innings_played = None
    catcher_caught_stealing = None
    catcher_stolen_bases_allowed = None
    catcher_earned_runs = None
    is_qualified_catcher = None
    is_qualified_pitcher = None
    successful_chances = None
    total_chances = None
    full_innings_played = None
    part_innings_played = None
    fielding_percentage = None
    range_factor = None
    zone_rating = None
    catcher_caught_stealing_percentage = None
    catcher_era = None
    def_warbr = None
    perfect_games = None
    wild_pitches = None
    third_innings = None
    team_earned_runs = None
    shutouts = None
    pickoff_attempts = None
    run_support = None
    pitches_as_starter = None
    quality_starts = None
    inherited_runners = None
    inherited_runners_scored = None
    opponent_total_bases = None
    is_qualified_saves = None
    full_innings = None
    part_innings = None
    blown_saves = None
    innings = None
    era = None
    whip = None
    caught_stealing_percentage = None
    pitches_per_start = None
    pitches_per_inning = None
    run_support_average = None
    opponent_average = None
    opponent_slug_average = None
    opponent_on_base_percentage = None
    opponent_ops = None
    save_percentage = None
    strikeouts_per_nine_innings = None
    strikeout_to_walk_ratio = None
    tough_losses = None
    cheap_wins = None
    save_opportunities_per_win = None
    runs_created = None
    batting_average = None
    pinch_average = None
    slug_average = None
    secondary_average = None
    on_base_percentage = None
    ops = None
    ground_to_fly_ratio = None
    runs_created_per_27_outs = None
    batter_rating = None
    at_bats_per_home_run = None
    stolen_base_percentage = None
    pitches_per_plate_appearance = None
    isolated_power = None
    walk_to_strikeout_ratio = None
    walks_per_plate_appearance = None
    secondary_average_minus_batting_average = None
    runs_produced = None
    runs_ratio = None
    patience_ratio = None
    balls_in_play_average = None
    mlb_rating = None
    offensive_wins_above_replacement = None
    wins_above_replacement = None
    earned_runs = None
    batters_hit = None
    sacrifice_bunts = None
    save_opportunities = None
    finishes = None
    balks = None
    batters_faced = None
    holds = None
    complete_games = None
    hit_by_pitch = None
    ground_balls = None
    strikeouts = None
    rbis = None
    sac_hits = None
    hits = None
    stolen_bases = None
    walks = None
    catcher_interference = None
    gidps = None
    sacrifice_flies = None
    at_bats = None
    home_runs = None
    grand_slam_home_runs = None
    runners_left_on_base = None
    triples = None
    game_winning_rbis = None
    intentional_walks = None
    doubles = None
    fly_balls = None
    caught_stealing = None
    pitches = None
    games_started = None
    pinch_at_bats = None
    pinch_hits = None
    player_rating = None
    is_qualified = None
    is_qualified_steals = None
    total_bases = None
    plate_appearances = None
    projected_home_runs = None
    extra_base_hits = None
    average_game_score = None
    average_field_goals_attempted = None
    average_three_point_field_goals_made = None
    average_three_point_field_goals_attempted = None
    average_free_throws_made = None
    average_free_throws_attempted = None
    average_points = None
    average_offensive_rebounds = None
    average_assists = None
    average_turnovers = None
    offensive_rebound_percentage = None
    estimated_possessions = None
    average_estimated_possessions = None
    points_per_estimated_possessions = None
    average_team_turnovers = None
    average_total_turnovers = None
    two_point_field_goals_made = None
    two_point_field_goals_attempted = None
    average_two_point_field_goals_made = None
    average_two_point_field_goals_attempted = None
    two_point_field_goal_percentage = None
    shooting_efficiency = None
    scoring_efficiency = None
    average_48_field_goals_made = None
    average_48_field_goals_attempted = None
    average_48_three_point_field_goals_made = None
    average_48_three_point_field_goals_attempted = None
    average_48_free_throws_made = None
    average_48_free_throws_attempted = None
    average_48_points = None
    average_48_offensive_rebounds = None
    average_48_assists = None
    average_48_turnovers = None
    average_rebounds = None
    average_fouls = None
    average_flagrant_fouls = None
    average_technical_fouls = None
    average_ejections = None
    average_disqualifications = None
    assist_turnover_ratio = None
    steal_foul_ratio = None
    block_foul_ratio = None
    average_team_rebounds = None
    total_technical_fouls = None
    team_assist_turnover_ratio = None
    steal_turnover_ratio = None
    average_48_rebounds = None
    average_48_fouls = None
    average_48_flagrant_fouls = None
    average_48_technical_fouls = None
    average_48_ejections = None
    average_48_disqualifications = None
    double_double = None
    triple_double = None
    field_goals_made = None
    free_throws_made = None
    three_point_percentage = None
    three_point_field_goals_made = None
    team_turnovers = None
    total_turnovers = None
    points_in_paint = None
    brick_index = None
    fast_break_points = None
    average_field_goals_made = None
    turnover_points = None
    average_defensive_rebounds = None
    average_blocks = None
    average_steals = None
    average_48_defensive_rebounds = None
    average_48_blocks = None
    average_48_steals = None
    largest_lead = None
    disqualifications = None
    flagrant_fouls = None
    fouls = None
    ejections = None
    technical_fouls = None
    vorp = None
    average_minutes = None
    nba_rating = None
    set_one_points = None
    set_two_points = None
    set_three_points = None
    set_four_points = None
    set_five_points = None
    for team_model in team_models:
        location = more_interesting(location, team_model.location)
        for player_model in team_model.players:
            player_id = player_model.identifier
            player_name_key = normalise_name(player_model.name)
            if player_model.identifier in player_identity_map:
                player_id = player_identity_map[player_id]
            elif player_name_key in names:
                player_id = names[player_name_key]
            else:
                names[player_name_key] = player_id
            players[player_id] = players.get(player_id, []) + [player_model]
        for odds_model in team_model.odds:
            key = f"{odds_model.bookie.identifier}-{odds_model.odds}"
            odds[key] = odds.get(key, []) + [odds_model]
        points = more_interesting(points, team_model.points)
        ladder_rank = more_interesting(ladder_rank, team_model.ladder_rank)
        for news_model in team_model.news:
            news_key = "-".join(
                [
                    news_model.title,
                    str(news_model.published),
                    news_model.summary,
                    news_model.source,
                ]
            )
            news[news_key] = news_model
        for social_model in team_model.social:
            social_key = "-".join(
                [social_model.network, social_model.post, str(social_model.published)]
            )
            social[social_key] = social_model
        field_goals = more_interesting(field_goals, team_model.field_goals)
        for coach_model in team_model.coaches:
            coach_id = coach_model.identifier
            coach_name_key = normalise_name(coach_model.name)
            if coach_name_key in coach_names:
                coach_id = coach_names[coach_name_key]
            else:
                coach_names[coach_name_key] = coach_id
            coaches[coach_id] = coaches.get(coach_id, []) + [coach_model]
        lbw = more_interesting(lbw, team_model.lbw)
        end_dt = more_interesting(end_dt, team_model.end_dt)
        runs = more_interesting(runs, team_model.runs)
        wickets = more_interesting(wickets, team_model.wickets)
        overs = more_interesting(overs, team_model.overs)
        balls = more_interesting(balls, team_model.balls)
        byes = more_interesting(byes, team_model.byes)
        leg_byes = more_interesting(leg_byes, team_model.leg_byes)
        wides = more_interesting(wides, team_model.wides)
        no_balls = more_interesting(no_balls, team_model.no_balls)
        penalties = more_interesting(penalties, team_model.penalties)
        balls_per_over = more_interesting(balls_per_over, team_model.balls_per_over)
        fours = more_interesting(fours, team_model.fours)
        sixes = more_interesting(sixes, team_model.sixes)
        catches = more_interesting(catches, team_model.catches)
        catches_dropped = more_interesting(catches_dropped, team_model.catches_dropped)
        disposal_efficiency = more_interesting(
            disposal_efficiency, team_model.disposal_efficiency
        )
        uncontested_marks = more_interesting(
            uncontested_marks, team_model.uncontested_marks
        )
        total_possessions = more_interesting(
            total_possessions, team_model.total_possessions
        )
        tackles_inside_50 = more_interesting(
            tackles_inside_50, team_model.tackles_inside_50
        )
        centre_clearances = more_interesting(
            centre_clearances, team_model.centre_clearances
        )
        stoppage_clearances = more_interesting(
            stoppage_clearances, team_model.stoppage_clearances
        )
        goal_accuracy = more_interesting(goal_accuracy, team_model.goal_accuracy)
        rushed_behinds = more_interesting(rushed_behinds, team_model.rushed_behinds)
        touched_behinds = more_interesting(touched_behinds, team_model.touched_behinds)
        left_behinds = more_interesting(left_behinds, team_model.left_behinds)
        left_posters = more_interesting(left_posters, team_model.left_posters)
        right_behinds = more_interesting(right_behinds, team_model.right_behinds)
        right_posters = more_interesting(right_posters, team_model.right_posters)
        total_interchange_count = more_interesting(
            total_interchange_count, team_model.total_interchange_count
        )
        interchange_count_q1 = more_interesting(
            interchange_count_q1, team_model.interchange_count_q1
        )
        interchange_count_q2 = more_interesting(
            interchange_count_q2, team_model.interchange_count_q2
        )
        interchange_count_q3 = more_interesting(
            interchange_count_q3, team_model.interchange_count_q3
        )
        interchange_count_q4 = more_interesting(
            interchange_count_q4, team_model.interchange_count_q4
        )
        game_winning_goals = more_interesting(
            game_winning_goals, team_model.game_winning_goals
        )
        headed_goals = more_interesting(headed_goals, team_model.headed_goals)
        inaccurate_crosses = more_interesting(
            inaccurate_crosses, team_model.inaccurate_crosses
        )
        inaccurate_long_balls = more_interesting(
            inaccurate_long_balls, team_model.inaccurate_long_balls
        )
        inaccurate_passes = more_interesting(
            inaccurate_passes, team_model.inaccurate_passes
        )
        inaccurate_through_balls = more_interesting(
            inaccurate_through_balls, team_model.inaccurate_through_balls
        )
        left_footed_shots = more_interesting(
            left_footed_shots, team_model.left_footed_shots
        )
        longball_percentage = more_interesting(
            longball_percentage, team_model.longball_percentage
        )
        offsides = more_interesting(offsides, team_model.offsides)
        penalty_kick_goals = more_interesting(
            penalty_kick_goals, team_model.penalty_kick_goals
        )
        penalty_kick_percentage = more_interesting(
            penalty_kick_percentage, team_model.penalty_kick_percentage
        )
        penalty_kick_shots = more_interesting(
            penalty_kick_shots, team_model.penalty_kick_shots
        )
        penalty_kicks_missed = more_interesting(
            penalty_kicks_missed, team_model.penalty_kicks_missed
        )
        possession_percentage = more_interesting(
            possession_percentage, team_model.possession_percentage
        )
        possession_time = more_interesting(possession_time, team_model.possession_time)
        right_footed_shots = more_interesting(
            right_footed_shots, team_model.right_footed_shots
        )
        shoot_out_goals = more_interesting(shoot_out_goals, team_model.shoot_out_goals)
        shoot_out_misses = more_interesting(
            shoot_out_misses, team_model.shoot_out_misses
        )
        shoot_out_percentage = more_interesting(
            shoot_out_percentage, team_model.shoot_out_percentage
        )
        shot_assists = more_interesting(shot_assists, team_model.shot_assists)
        shot_percentage = more_interesting(shot_percentage, team_model.shot_percentage)
        shots_headed = more_interesting(shots_headed, team_model.shots_headed)
        shots_off_target = more_interesting(
            shots_off_target, team_model.shots_off_target
        )
        shots_on_post = more_interesting(shots_on_post, team_model.shots_on_post)
        shots_on_target = more_interesting(shots_on_target, team_model.shots_on_target)
        through_ball_percentage = more_interesting(
            through_ball_percentage, team_model.through_ball_percentage
        )
        total_crosses = more_interesting(total_crosses, team_model.total_crosses)
        total_goals = more_interesting(total_goals, team_model.total_goals)
        total_long_balls = more_interesting(
            total_long_balls, team_model.total_long_balls
        )
        total_passes = more_interesting(total_passes, team_model.total_passes)
        total_shots = more_interesting(total_shots, team_model.total_shots)
        total_through_balls = more_interesting(
            total_through_balls, team_model.total_through_balls
        )
        draws = more_interesting(draws, team_model.draws)
        sub_outs = more_interesting(sub_outs, team_model.sub_outs)
        suspensions = more_interesting(suspensions, team_model.suspensions)
        time_ended = more_interesting(time_ended, team_model.time_ended)
        time_started = more_interesting(time_started, team_model.time_started)
        win_percentage = more_interesting(win_percentage, team_model.win_percentage)
        wins = more_interesting(wins, team_model.wins)
        won_corners = more_interesting(won_corners, team_model.won_corners)
        yellow_cards = more_interesting(yellow_cards, team_model.yellow_cards)
        clean_sheet = more_interesting(clean_sheet, team_model.clean_sheet)
        crosses_caught = more_interesting(crosses_caught, team_model.crosses_caught)
        goals_conceded = more_interesting(goals_conceded, team_model.goals_conceded)
        partial_clean_sheet = more_interesting(
            partial_clean_sheet, team_model.partial_clean_sheet
        )
        penalty_kick_conceded = more_interesting(
            penalty_kick_conceded, team_model.penalty_kick_conceded
        )
        penalty_kick_save_percentage = more_interesting(
            penalty_kick_save_percentage, team_model.penalty_kick_save_percentage
        )
        penalty_kicks_faced = more_interesting(
            penalty_kicks_faced, team_model.penalty_kicks_faced
        )
        penalty_kicks_saved = more_interesting(
            penalty_kicks_saved, team_model.penalty_kicks_saved
        )
        punches = more_interesting(punches, team_model.punches)
        saves = more_interesting(saves, team_model.saves)
        shoot_out_kicks_faced = more_interesting(
            shoot_out_kicks_faced, team_model.shoot_out_kicks_faced
        )
        shoot_out_kicks_saved = more_interesting(
            shoot_out_kicks_saved, team_model.shoot_out_kicks_saved
        )
        shoot_out_save_percentage = more_interesting(
            shoot_out_save_percentage, team_model.shoot_out_save_percentage
        )
        shots_faced = more_interesting(shots_faced, team_model.shots_faced)
        smothers = more_interesting(smothers, team_model.smothers)
        unclaimed_crosses = more_interesting(
            unclaimed_crosses, team_model.unclaimed_crosses
        )
        accurate_crosses = more_interesting(
            accurate_crosses, team_model.accurate_crosses
        )
        accurate_long_balls = more_interesting(
            accurate_long_balls, team_model.accurate_long_balls
        )
        accurate_passes = more_interesting(accurate_passes, team_model.accurate_passes)
        accurate_through_balls = more_interesting(
            accurate_through_balls, team_model.accurate_through_balls
        )
        cross_percentage = more_interesting(
            cross_percentage, team_model.cross_percentage
        )
        free_kick_goals = more_interesting(free_kick_goals, team_model.free_kick_goals)
        free_kick_percentage = more_interesting(
            free_kick_percentage, team_model.free_kick_percentage
        )
        free_kick_shots = more_interesting(free_kick_shots, team_model.free_kick_shots)
        game_winning_assists = more_interesting(
            game_winning_assists, team_model.game_winning_assists
        )
        blocked_shots = more_interesting(blocked_shots, team_model.blocked_shots)
        effective_clearances = more_interesting(
            effective_clearances, team_model.effective_clearances
        )
        effective_tackles = more_interesting(
            effective_tackles, team_model.effective_tackles
        )
        ineffective_tackles = more_interesting(
            ineffective_tackles, team_model.ineffective_tackles
        )
        interceptions = more_interesting(interceptions, team_model.interceptions)
        tackle_percentage = more_interesting(
            tackle_percentage, team_model.tackle_percentage
        )
        appearances = more_interesting(appearances, team_model.appearances)
        average_rating_from_correspondent = more_interesting(
            average_rating_from_correspondent,
            team_model.average_rating_from_correspondent,
        )
        average_rating_from_data_feed = more_interesting(
            average_rating_from_data_feed, team_model.average_rating_from_data_feed
        )
        average_rating_from_editor = more_interesting(
            average_rating_from_editor, team_model.average_rating_from_editor
        )
        average_rating_from_user = more_interesting(
            average_rating_from_user, team_model.average_rating_from_user
        )
        did_not_play = more_interesting(did_not_play, team_model.did_not_play)
        fouls_committed = more_interesting(fouls_committed, team_model.fouls_committed)
        fouls_suffered = more_interesting(fouls_suffered, team_model.fouls_suffered)
        goal_difference = more_interesting(goal_difference, team_model.goal_difference)
        losses = more_interesting(losses, team_model.losses)
        lost_corners = more_interesting(lost_corners, team_model.lost_corners)
        minutes = more_interesting(minutes, team_model.minutes)
        own_goals = more_interesting(own_goals, team_model.own_goals)
        pass_percentage = more_interesting(pass_percentage, team_model.pass_percentage)
        red_cards = more_interesting(red_cards, team_model.red_cards)
        starts = more_interesting(starts, team_model.starts)
        sub_ins = more_interesting(sub_ins, team_model.sub_ins)
        pitch_count = more_interesting(pitch_count, team_model.pitch_count)
        strikes = more_interesting(strikes, team_model.strikes)
        strike_pitch_ratio = more_interesting(
            strike_pitch_ratio, team_model.strike_pitch_ratio
        )
        games_played = more_interesting(games_played, team_model.games_played)
        team_games_played = more_interesting(
            team_games_played, team_model.team_games_played
        )
        double_plays = more_interesting(double_plays, team_model.double_plays)
        opportunities = more_interesting(opportunities, team_model.opportunities)
        errors = more_interesting(errors, team_model.errors)
        passed_balls = more_interesting(passed_balls, team_model.passed_balls)
        outfield_assists = more_interesting(
            outfield_assists, team_model.outfield_assists
        )
        pickoffs = more_interesting(pickoffs, team_model.pickoffs)
        putouts = more_interesting(putouts, team_model.putouts)
        outs_on_field = more_interesting(outs_on_field, team_model.outs_on_field)
        triple_plays = more_interesting(triple_plays, team_model.triple_plays)
        balls_in_zone = more_interesting(balls_in_zone, team_model.balls_in_zone)
        extra_bases = more_interesting(extra_bases, team_model.extra_bases)
        outs_made = more_interesting(outs_made, team_model.outs_made)
        catcher_third_innings_played = more_interesting(
            catcher_third_innings_played, team_model.catcher_third_innings_played
        )
        catcher_caught_stealing = more_interesting(
            catcher_caught_stealing, team_model.catcher_caught_stealing
        )
        catcher_stolen_bases_allowed = more_interesting(
            catcher_stolen_bases_allowed, team_model.catcher_stolen_bases_allowed
        )
        catcher_earned_runs = more_interesting(
            catcher_earned_runs, team_model.catcher_earned_runs
        )
        is_qualified_catcher = more_interesting(
            is_qualified_catcher, team_model.is_qualified_catcher
        )
        is_qualified_pitcher = more_interesting(
            is_qualified_pitcher, team_model.is_qualified_pitcher
        )
        successful_chances = more_interesting(
            successful_chances, team_model.successful_chances
        )
        total_chances = more_interesting(total_chances, team_model.total_chances)
        full_innings_played = more_interesting(
            full_innings_played, team_model.full_innings_played
        )
        part_innings_played = more_interesting(
            part_innings_played, team_model.part_innings_played
        )
        fielding_percentage = more_interesting(
            fielding_percentage, team_model.fielding_percentage
        )
        range_factor = more_interesting(range_factor, team_model.range_factor)
        zone_rating = more_interesting(zone_rating, team_model.zone_rating)
        catcher_caught_stealing_percentage = more_interesting(
            catcher_caught_stealing_percentage,
            team_model.catcher_caught_stealing_percentage,
        )
        catcher_era = more_interesting(catcher_era, team_model.catcher_era)
        def_warbr = more_interesting(def_warbr, team_model.def_warbr)
        perfect_games = more_interesting(perfect_games, team_model.perfect_games)
        wild_pitches = more_interesting(wild_pitches, team_model.wild_pitches)
        third_innings = more_interesting(third_innings, team_model.third_innings)
        team_earned_runs = more_interesting(
            team_earned_runs, team_model.team_earned_runs
        )
        shutouts = more_interesting(shutouts, team_model.shutouts)
        pickoff_attempts = more_interesting(
            pickoff_attempts, team_model.pickoff_attempts
        )
        run_support = more_interesting(run_support, team_model.run_support)
        pitches_as_starter = more_interesting(
            pitches_as_starter, team_model.pitches_as_starter
        )
        quality_starts = more_interesting(quality_starts, team_model.quality_starts)
        inherited_runners = more_interesting(
            inherited_runners, team_model.inherited_runners
        )
        inherited_runners_scored = more_interesting(
            inherited_runners_scored, team_model.inherited_runners_scored
        )
        opponent_total_bases = more_interesting(
            opponent_total_bases, team_model.opponent_total_bases
        )
        is_qualified_saves = more_interesting(
            is_qualified_saves, team_model.is_qualified_saves
        )
        full_innings = more_interesting(full_innings, team_model.full_innings)
        part_innings = more_interesting(part_innings, team_model.part_innings)
        blown_saves = more_interesting(blown_saves, team_model.blown_saves)
        innings = more_interesting(innings, team_model.innings)
        era = more_interesting(era, team_model.era)
        whip = more_interesting(whip, team_model.whip)
        caught_stealing_percentage = more_interesting(
            caught_stealing_percentage, team_model.caught_stealing_percentage
        )
        pitches_per_start = more_interesting(
            pitches_per_start, team_model.pitches_per_start
        )
        pitches_per_inning = more_interesting(
            pitches_per_inning, team_model.pitches_per_inning
        )
        run_support_average = more_interesting(
            run_support_average, team_model.run_support_average
        )
        opponent_average = more_interesting(
            opponent_average, team_model.opponent_average
        )
        opponent_slug_average = more_interesting(
            opponent_slug_average, team_model.opponent_slug_average
        )
        opponent_on_base_percentage = more_interesting(
            opponent_on_base_percentage, team_model.opponent_on_base_percentage
        )
        opponent_ops = more_interesting(opponent_ops, team_model.opponent_ops)
        save_percentage = more_interesting(save_percentage, team_model.save_percentage)
        strikeouts_per_nine_innings = more_interesting(
            strikeouts_per_nine_innings, team_model.strikeouts_per_nine_innings
        )
        strikeout_to_walk_ratio = more_interesting(
            strikeout_to_walk_ratio, team_model.strikeout_to_walk_ratio
        )
        tough_losses = more_interesting(tough_losses, team_model.tough_losses)
        cheap_wins = more_interesting(cheap_wins, team_model.cheap_wins)
        save_opportunities_per_win = more_interesting(
            save_opportunities_per_win, team_model.save_opportunities_per_win
        )
        runs_created = more_interesting(runs_created, team_model.runs_created)
        batting_average = more_interesting(batting_average, team_model.batting_average)
        pinch_average = more_interesting(pinch_average, team_model.pinch_average)
        slug_average = more_interesting(slug_average, team_model.slug_average)
        secondary_average = more_interesting(
            secondary_average, team_model.secondary_average
        )
        on_base_percentage = more_interesting(
            on_base_percentage, team_model.on_base_percentage
        )
        ops = more_interesting(ops, team_model.ops)
        ground_to_fly_ratio = more_interesting(
            ground_to_fly_ratio, team_model.ground_to_fly_ratio
        )
        runs_created_per_27_outs = more_interesting(
            runs_created_per_27_outs, team_model.runs_created_per_27_outs
        )
        batter_rating = more_interesting(batter_rating, team_model.batter_rating)
        at_bats_per_home_run = more_interesting(
            at_bats_per_home_run, team_model.at_bats_per_home_run
        )
        stolen_base_percentage = more_interesting(
            stolen_base_percentage, team_model.stolen_base_percentage
        )
        pitches_per_plate_appearance = more_interesting(
            pitches_per_plate_appearance, team_model.pitches_per_plate_appearance
        )
        isolated_power = more_interesting(isolated_power, team_model.isolated_power)
        walk_to_strikeout_ratio = more_interesting(
            walk_to_strikeout_ratio, team_model.walk_to_strikeout_ratio
        )
        walks_per_plate_appearance = more_interesting(
            walks_per_plate_appearance, team_model.walks_per_plate_appearance
        )
        secondary_average_minus_batting_average = more_interesting(
            secondary_average_minus_batting_average,
            team_model.secondary_average_minus_batting_average,
        )
        runs_produced = more_interesting(runs_produced, team_model.runs_produced)
        runs_ratio = more_interesting(runs_ratio, team_model.runs_ratio)
        patience_ratio = more_interesting(patience_ratio, team_model.patience_ratio)
        balls_in_play_average = more_interesting(
            balls_in_play_average, team_model.balls_in_play_average
        )
        mlb_rating = more_interesting(mlb_rating, team_model.mlb_rating)
        offensive_wins_above_replacement = more_interesting(
            offensive_wins_above_replacement,
            team_model.offensive_wins_above_replacement,
        )
        wins_above_replacement = more_interesting(
            wins_above_replacement, team_model.wins_above_replacement
        )
        earned_runs = more_interesting(earned_runs, team_model.earned_runs)
        batters_hit = more_interesting(batters_hit, team_model.batters_hit)
        sacrifice_bunts = more_interesting(sacrifice_bunts, team_model.sacrifice_bunts)
        save_opportunities = more_interesting(
            save_opportunities, team_model.save_opportunities
        )
        finishes = more_interesting(finishes, team_model.finishes)
        balks = more_interesting(balks, team_model.balks)
        batters_faced = more_interesting(batters_faced, team_model.batters_faced)
        holds = more_interesting(holds, team_model.holds)
        complete_games = more_interesting(complete_games, team_model.complete_games)
        hit_by_pitch = more_interesting(hit_by_pitch, team_model.hit_by_pitch)
        ground_balls = more_interesting(ground_balls, team_model.ground_balls)
        strikeouts = more_interesting(strikeouts, team_model.strikeouts)
        rbis = more_interesting(rbis, team_model.rbis)
        sac_hits = more_interesting(sac_hits, team_model.sac_hits)
        hits = more_interesting(hits, team_model.hits)
        stolen_bases = more_interesting(stolen_bases, team_model.stolen_bases)
        walks = more_interesting(walks, team_model.walks)
        catcher_interference = more_interesting(
            catcher_interference, team_model.catcher_interference
        )
        gidps = more_interesting(gidps, team_model.gidps)
        sacrifice_flies = more_interesting(sacrifice_flies, team_model.sacrifice_flies)
        at_bats = more_interesting(at_bats, team_model.at_bats)
        home_runs = more_interesting(home_runs, team_model.home_runs)
        grand_slam_home_runs = more_interesting(
            grand_slam_home_runs, team_model.grand_slam_home_runs
        )
        runners_left_on_base = more_interesting(
            runners_left_on_base, team_model.runners_left_on_base
        )
        triples = more_interesting(triples, team_model.triples)
        game_winning_rbis = more_interesting(
            game_winning_rbis, team_model.game_winning_rbis
        )
        intentional_walks = more_interesting(
            intentional_walks, team_model.intentional_walks
        )
        doubles = more_interesting(doubles, team_model.doubles)
        fly_balls = more_interesting(fly_balls, team_model.fly_balls)
        caught_stealing = more_interesting(caught_stealing, team_model.caught_stealing)
        pitches = more_interesting(pitches, team_model.pitches)
        games_started = more_interesting(games_started, team_model.games_started)
        pinch_at_bats = more_interesting(pinch_at_bats, team_model.pinch_at_bats)
        pinch_hits = more_interesting(pinch_hits, team_model.pinch_hits)
        player_rating = more_interesting(player_rating, team_model.player_rating)
        is_qualified = more_interesting(is_qualified, team_model.is_qualified)
        is_qualified_steals = more_interesting(
            is_qualified_steals, team_model.is_qualified_steals
        )
        total_bases = more_interesting(total_bases, team_model.total_bases)
        plate_appearances = more_interesting(
            plate_appearances, team_model.plate_appearances
        )
        projected_home_runs = more_interesting(
            projected_home_runs, team_model.projected_home_runs
        )
        extra_base_hits = more_interesting(extra_base_hits, team_model.extra_base_hits)
        average_game_score = more_interesting(
            average_game_score, team_model.average_game_score
        )
        average_field_goals_attempted = more_interesting(
            average_field_goals_attempted, team_model.average_field_goals_attempted
        )
        average_three_point_field_goals_made = more_interesting(
            average_three_point_field_goals_made,
            team_model.average_three_point_field_goals_made,
        )
        average_three_point_field_goals_attempted = more_interesting(
            average_three_point_field_goals_attempted,
            team_model.average_three_point_field_goals_attempted,
        )
        average_free_throws_made = more_interesting(
            average_free_throws_made, team_model.average_free_throws_made
        )
        average_free_throws_attempted = more_interesting(
            average_free_throws_attempted, team_model.average_free_throws_attempted
        )
        average_points = more_interesting(average_points, team_model.average_points)
        average_offensive_rebounds = more_interesting(
            average_offensive_rebounds, team_model.average_offensive_rebounds
        )
        average_assists = more_interesting(average_assists, team_model.average_assists)
        average_turnovers = more_interesting(
            average_turnovers, team_model.average_turnovers
        )
        offensive_rebound_percentage = more_interesting(
            offensive_rebound_percentage, team_model.offensive_rebound_percentage
        )
        estimated_possessions = more_interesting(
            estimated_possessions, team_model.estimated_possessions
        )
        average_estimated_possessions = more_interesting(
            average_estimated_possessions, team_model.average_estimated_possessions
        )
        points_per_estimated_possessions = more_interesting(
            points_per_estimated_possessions,
            team_model.points_per_estimated_possessions,
        )
        average_team_turnovers = more_interesting(
            average_team_turnovers, team_model.average_team_turnovers
        )
        average_total_turnovers = more_interesting(
            average_total_turnovers, team_model.average_total_turnovers
        )
        two_point_field_goals_made = more_interesting(
            two_point_field_goals_made, team_model.two_point_field_goals_made
        )
        two_point_field_goals_attempted = more_interesting(
            two_point_field_goals_attempted, team_model.two_point_field_goals_attempted
        )
        average_two_point_field_goals_made = more_interesting(
            average_two_point_field_goals_made,
            team_model.average_two_point_field_goals_made,
        )
        average_two_point_field_goals_attempted = more_interesting(
            average_two_point_field_goals_attempted,
            team_model.average_two_point_field_goals_attempted,
        )
        two_point_field_goal_percentage = more_interesting(
            two_point_field_goal_percentage, team_model.two_point_field_goal_percentage
        )
        shooting_efficiency = more_interesting(
            shooting_efficiency, team_model.shooting_efficiency
        )
        scoring_efficiency = more_interesting(
            scoring_efficiency, team_model.scoring_efficiency
        )
        average_48_field_goals_made = more_interesting(
            average_48_field_goals_made, team_model.average_48_field_goals_made
        )
        average_48_field_goals_attempted = more_interesting(
            average_48_field_goals_attempted,
            team_model.average_48_field_goals_attempted,
        )
        average_48_three_point_field_goals_made = more_interesting(
            average_48_three_point_field_goals_made,
            team_model.average_48_three_point_field_goals_made,
        )
        average_48_three_point_field_goals_attempted = more_interesting(
            average_48_three_point_field_goals_attempted,
            team_model.average_48_three_point_field_goals_attempted,
        )
        average_48_free_throws_made = more_interesting(
            average_48_free_throws_made, team_model.average_48_free_throws_made
        )
        average_48_free_throws_attempted = more_interesting(
            average_48_free_throws_attempted,
            team_model.average_48_free_throws_attempted,
        )
        average_48_points = more_interesting(
            average_48_points, team_model.average_48_points
        )
        average_48_offensive_rebounds = more_interesting(
            average_48_offensive_rebounds, team_model.average_48_offensive_rebounds
        )
        average_48_assists = more_interesting(
            average_48_assists, team_model.average_48_assists
        )
        average_48_turnovers = more_interesting(
            average_48_turnovers, team_model.average_48_turnovers
        )
        average_rebounds = more_interesting(
            average_rebounds, team_model.average_rebounds
        )
        average_fouls = more_interesting(average_fouls, team_model.average_fouls)
        average_flagrant_fouls = more_interesting(
            average_flagrant_fouls, team_model.average_flagrant_fouls
        )
        average_technical_fouls = more_interesting(
            average_technical_fouls, team_model.average_technical_fouls
        )
        average_ejections = more_interesting(
            average_ejections, team_model.average_ejections
        )
        average_disqualifications = more_interesting(
            average_disqualifications, team_model.average_disqualifications
        )
        assist_turnover_ratio = more_interesting(
            assist_turnover_ratio, team_model.assist_turnover_ratio
        )
        steal_foul_ratio = more_interesting(
            steal_foul_ratio, team_model.steal_foul_ratio
        )
        block_foul_ratio = more_interesting(
            block_foul_ratio, team_model.block_foul_ratio
        )
        average_team_rebounds = more_interesting(
            average_team_rebounds, team_model.average_team_rebounds
        )
        total_technical_fouls = more_interesting(
            total_technical_fouls, team_model.total_technical_fouls
        )
        team_assist_turnover_ratio = more_interesting(
            team_assist_turnover_ratio, team_model.team_assist_turnover_ratio
        )
        steal_turnover_ratio = more_interesting(
            steal_turnover_ratio, team_model.steal_turnover_ratio
        )
        average_48_rebounds = more_interesting(
            average_48_rebounds, team_model.average_48_rebounds
        )
        average_48_fouls = more_interesting(
            average_48_fouls, team_model.average_48_fouls
        )
        average_48_flagrant_fouls = more_interesting(
            average_48_flagrant_fouls, team_model.average_48_flagrant_fouls
        )
        average_48_technical_fouls = more_interesting(
            average_48_technical_fouls, team_model.average_48_technical_fouls
        )
        average_48_ejections = more_interesting(
            average_48_ejections, team_model.average_48_ejections
        )
        average_48_disqualifications = more_interesting(
            average_48_disqualifications, team_model.average_48_disqualifications
        )
        double_double = more_interesting(double_double, team_model.double_double)
        triple_double = more_interesting(triple_double, team_model.triple_double)
        field_goals_made = more_interesting(
            field_goals_made, team_model.field_goals_made
        )
        free_throws_made = more_interesting(
            free_throws_made, team_model.free_throws_made
        )
        three_point_percentage = more_interesting(
            three_point_percentage, team_model.three_point_percentage
        )
        three_point_field_goals_made = more_interesting(
            three_point_field_goals_made, team_model.three_point_field_goals_made
        )
        team_turnovers = more_interesting(team_turnovers, team_model.team_turnovers)
        total_turnovers = more_interesting(total_turnovers, team_model.total_turnovers)
        points_in_paint = more_interesting(points_in_paint, team_model.points_in_paint)
        brick_index = more_interesting(brick_index, team_model.brick_index)
        fast_break_points = more_interesting(
            fast_break_points, team_model.fast_break_points
        )
        average_field_goals_made = more_interesting(
            average_field_goals_made, team_model.average_field_goals_made
        )
        turnover_points = more_interesting(turnover_points, team_model.turnover_points)
        average_defensive_rebounds = more_interesting(
            average_defensive_rebounds, team_model.average_defensive_rebounds
        )
        average_blocks = more_interesting(average_blocks, team_model.average_blocks)
        average_steals = more_interesting(average_steals, team_model.average_steals)
        average_48_defensive_rebounds = more_interesting(
            average_48_defensive_rebounds, team_model.average_48_defensive_rebounds
        )
        average_48_blocks = more_interesting(
            average_48_blocks, team_model.average_48_blocks
        )
        average_48_steals = more_interesting(
            average_48_steals, team_model.average_48_steals
        )
        largest_lead = more_interesting(largest_lead, team_model.largest_lead)
        disqualifications = more_interesting(
            disqualifications, team_model.disqualifications
        )
        flagrant_fouls = more_interesting(flagrant_fouls, team_model.flagrant_fouls)
        fouls = more_interesting(fouls, team_model.fouls)
        ejections = more_interesting(ejections, team_model.ejections)
        technical_fouls = more_interesting(technical_fouls, team_model.technical_fouls)
        vorp = more_interesting(vorp, team_model.vorp)
        average_minutes = more_interesting(average_minutes, team_model.average_minutes)
        nba_rating = more_interesting(nba_rating, team_model.nba_rating)
        set_one_points = more_interesting(set_one_points, team_model.set_one_points)
        set_two_points = more_interesting(set_two_points, team_model.set_two_points)
        set_three_points = more_interesting(
            set_three_points, team_model.set_three_points
        )
        set_four_points = more_interesting(set_four_points, team_model.set_four_points)
        set_five_points = more_interesting(set_five_points, team_model.set_five_points)

    player_list = [
        create_combined_player_model(v, k, player_ffill) for k, v in players.items()
    ]
    player_list.sort(key=functools.cmp_to_key(_compare_player_models))

    team_model = TeamModel.model_construct(
        identifier=identifier,
        name=team_models[0].name,
        location=location,
        players=player_list,
        odds=[x[0] for x in odds.values()],
        points=points,
        ladder_rank=ladder_rank,
        news=sorted(news.values(), key=lambda x: x.published),
        social=sorted(social.values(), key=lambda x: x.published),
        field_goals=field_goals,
        coaches=[
            create_combined_coach_model(v, k, coach_ffill) for k, v in coaches.items()
        ],
        lbw=lbw,
        end_dt=end_dt,
        runs=runs,
        wickets=wickets,
        overs=overs,
        balls=balls,
        byes=byes,
        leg_byes=leg_byes,
        wides=wides,
        no_balls=no_balls,
        penalties=penalties,
        balls_per_over=balls_per_over,
        fours=fours,
        sixes=sixes,
        catches=catches,
        catches_dropped=catches_dropped,
        disposal_efficiency=disposal_efficiency,
        uncontested_marks=uncontested_marks,
        total_possessions=total_possessions,
        tackles_inside_50=tackles_inside_50,
        centre_clearances=centre_clearances,
        stoppage_clearances=stoppage_clearances,
        goal_accuracy=goal_accuracy,
        rushed_behinds=rushed_behinds,
        touched_behinds=touched_behinds,
        left_behinds=left_behinds,
        left_posters=left_posters,
        right_behinds=right_behinds,
        right_posters=right_posters,
        total_interchange_count=total_interchange_count,
        interchange_count_q1=interchange_count_q1,
        interchange_count_q2=interchange_count_q2,
        interchange_count_q3=interchange_count_q3,
        interchange_count_q4=interchange_count_q4,
        game_winning_goals=game_winning_goals,
        headed_goals=headed_goals,
        inaccurate_crosses=inaccurate_crosses,
        inaccurate_long_balls=inaccurate_long_balls,
        inaccurate_passes=inaccurate_passes,
        inaccurate_through_balls=inaccurate_through_balls,
        left_footed_shots=left_footed_shots,
        longball_percentage=longball_percentage,
        offsides=offsides,
        penalty_kick_goals=penalty_kick_goals,
        penalty_kick_percentage=penalty_kick_percentage,
        penalty_kick_shots=penalty_kick_shots,
        penalty_kicks_missed=penalty_kicks_missed,
        possession_percentage=possession_percentage,
        possession_time=possession_time,
        right_footed_shots=right_footed_shots,
        shoot_out_goals=shoot_out_goals,
        shoot_out_misses=shoot_out_misses,
        shoot_out_percentage=shoot_out_percentage,
        shot_assists=shot_assists,
        shot_percentage=shot_percentage,
        shots_headed=shots_headed,
        shots_off_target=shots_off_target,
        shots_on_post=shots_on_post,
        shots_on_target=shots_on_target,
        through_ball_percentage=through_ball_percentage,
        total_crosses=total_crosses,
        total_goals=total_goals,
        total_long_balls=total_long_balls,
        total_passes=total_passes,
        total_shots=total_shots,
        total_through_balls=total_through_balls,
        draws=draws,
        sub_outs=sub_outs,
        suspensions=suspensions,
        time_ended=time_ended,
        time_started=time_started,
        win_percentage=win_percentage,
        wins=wins,
        won_corners=won_corners,
        yellow_cards=yellow_cards,
        clean_sheet=clean_sheet,
        crosses_caught=crosses_caught,
        goals_conceded=goals_conceded,
        partial_clean_sheet=partial_clean_sheet,
        penalty_kick_conceded=penalty_kick_conceded,
        penalty_kick_save_percentage=penalty_kick_save_percentage,
        penalty_kicks_faced=penalty_kicks_faced,
        penalty_kicks_saved=penalty_kicks_saved,
        punches=punches,
        saves=saves,
        shoot_out_kicks_faced=shoot_out_kicks_faced,
        shoot_out_kicks_saved=shoot_out_kicks_saved,
        shoot_out_save_percentage=shoot_out_save_percentage,
        shots_faced=shots_faced,
        smothers=smothers,
        unclaimed_crosses=unclaimed_crosses,
        accurate_crosses=accurate_crosses,
        accurate_long_balls=accurate_long_balls,
        accurate_passes=accurate_passes,
        accurate_through_balls=accurate_through_balls,
        cross_percentage=cross_percentage,
        free_kick_goals=free_kick_goals,
        free_kick_percentage=free_kick_percentage,
        free_kick_shots=free_kick_shots,
        game_winning_assists=game_winning_assists,
        blocked_shots=blocked_shots,
        effective_clearances=effective_clearances,
        effective_tackles=effective_tackles,
        ineffective_tackles=ineffective_tackles,
        interceptions=interceptions,
        tackle_percentage=tackle_percentage,
        appearances=appearances,
        average_rating_from_correspondent=average_rating_from_correspondent,
        average_rating_from_data_feed=average_rating_from_data_feed,
        average_rating_from_editor=average_rating_from_editor,
        average_rating_from_user=average_rating_from_user,
        did_not_play=did_not_play,
        fouls_committed=fouls_committed,
        fouls_suffered=fouls_suffered,
        goal_difference=goal_difference,
        losses=losses,
        lost_corners=lost_corners,
        minutes=minutes,
        own_goals=own_goals,
        pass_percentage=pass_percentage,
        red_cards=red_cards,
        starts=starts,
        sub_ins=sub_ins,
        pitch_count=pitch_count,
        strikes=strikes,
        strike_pitch_ratio=strike_pitch_ratio,
        games_played=games_played,
        team_games_played=team_games_played,
        double_plays=double_plays,
        opportunities=opportunities,
        errors=errors,
        passed_balls=passed_balls,
        outfield_assists=outfield_assists,
        pickoffs=pickoffs,
        putouts=putouts,
        outs_on_field=outs_on_field,
        triple_plays=triple_plays,
        balls_in_zone=balls_in_zone,
        extra_bases=extra_bases,
        outs_made=outs_made,
        catcher_third_innings_played=catcher_third_innings_played,
        catcher_caught_stealing=catcher_caught_stealing,
        catcher_stolen_bases_allowed=catcher_stolen_bases_allowed,
        catcher_earned_runs=catcher_earned_runs,
        is_qualified_catcher=is_qualified_catcher,
        is_qualified_pitcher=is_qualified_pitcher,
        successful_chances=successful_chances,
        total_chances=total_chances,
        full_innings_played=full_innings_played,
        part_innings_played=part_innings_played,
        fielding_percentage=fielding_percentage,
        range_factor=range_factor,
        zone_rating=zone_rating,
        catcher_caught_stealing_percentage=catcher_caught_stealing_percentage,
        catcher_era=catcher_era,
        def_warbr=def_warbr,
        perfect_games=perfect_games,
        wild_pitches=wild_pitches,
        third_innings=third_innings,
        team_earned_runs=team_earned_runs,
        shutouts=shutouts,
        pickoff_attempts=pickoff_attempts,
        run_support=run_support,
        pitches_as_starter=pitches_as_starter,
        quality_starts=quality_starts,
        inherited_runners=inherited_runners,
        inherited_runners_scored=inherited_runners_scored,
        opponent_total_bases=opponent_total_bases,
        is_qualified_saves=is_qualified_saves,
        full_innings=full_innings,
        part_innings=part_innings,
        blown_saves=blown_saves,
        innings=innings,
        era=era,
        whip=whip,
        caught_stealing_percentage=caught_stealing_percentage,
        pitches_per_start=pitches_per_start,
        pitches_per_inning=pitches_per_inning,
        run_support_average=run_support_average,
        opponent_average=opponent_average,
        opponent_slug_average=opponent_slug_average,
        opponent_on_base_percentage=opponent_on_base_percentage,
        opponent_ops=opponent_ops,
        save_percentage=save_percentage,
        strikeouts_per_nine_innings=strikeouts_per_nine_innings,
        strikeout_to_walk_ratio=strikeout_to_walk_ratio,
        tough_losses=tough_losses,
        cheap_wins=cheap_wins,
        save_opportunities_per_win=save_opportunities_per_win,
        runs_created=runs_created,
        batting_average=batting_average,
        pinch_average=pinch_average,
        slug_average=slug_average,
        secondary_average=secondary_average,
        on_base_percentage=on_base_percentage,
        ops=ops,
        ground_to_fly_ratio=ground_to_fly_ratio,
        runs_created_per_27_outs=runs_created_per_27_outs,
        batter_rating=batter_rating,
        at_bats_per_home_run=at_bats_per_home_run,
        stolen_base_percentage=stolen_base_percentage,
        pitches_per_plate_appearance=pitches_per_plate_appearance,
        isolated_power=isolated_power,
        walk_to_strikeout_ratio=walk_to_strikeout_ratio,
        walks_per_plate_appearance=walks_per_plate_appearance,
        secondary_average_minus_batting_average=secondary_average_minus_batting_average,
        runs_produced=runs_produced,
        runs_ratio=runs_ratio,
        patience_ratio=patience_ratio,
        balls_in_play_average=balls_in_play_average,
        mlb_rating=mlb_rating,
        offensive_wins_above_replacement=offensive_wins_above_replacement,
        wins_above_replacement=wins_above_replacement,
        earned_runs=earned_runs,
        batters_hit=batters_hit,
        sacrifice_bunts=sacrifice_bunts,
        save_opportunities=save_opportunities,
        finishes=finishes,
        balks=balks,
        batters_faced=batters_faced,
        holds=holds,
        complete_games=complete_games,
        hit_by_pitch=hit_by_pitch,
        ground_balls=ground_balls,
        strikeouts=strikeouts,
        rbis=rbis,
        sac_hits=sac_hits,
        hits=hits,
        stolen_bases=stolen_bases,
        walks=walks,
        catcher_interference=catcher_interference,
        gidps=gidps,
        sacrifice_flies=sacrifice_flies,
        at_bats=at_bats,
        home_runs=home_runs,
        grand_slam_home_runs=grand_slam_home_runs,
        runners_left_on_base=runners_left_on_base,
        triples=triples,
        game_winning_rbis=game_winning_rbis,
        intentional_walks=intentional_walks,
        doubles=doubles,
        fly_balls=fly_balls,
        caught_stealing=caught_stealing,
        pitches=pitches,
        games_started=games_started,
        pinch_at_bats=pinch_at_bats,
        pinch_hits=pinch_hits,
        player_rating=player_rating,
        is_qualified=is_qualified,
        is_qualified_steals=is_qualified_steals,
        total_bases=total_bases,
        plate_appearances=plate_appearances,
        projected_home_runs=projected_home_runs,
        extra_base_hits=extra_base_hits,
        average_game_score=average_game_score,
        average_field_goals_attempted=average_field_goals_attempted,
        average_three_point_field_goals_made=average_three_point_field_goals_made,
        average_three_point_field_goals_attempted=average_three_point_field_goals_attempted,
        average_free_throws_made=average_free_throws_made,
        average_free_throws_attempted=average_free_throws_attempted,
        average_points=average_points,
        average_offensive_rebounds=average_offensive_rebounds,
        average_assists=average_assists,
        average_turnovers=average_turnovers,
        offensive_rebound_percentage=offensive_rebound_percentage,
        estimated_possessions=estimated_possessions,
        average_estimated_possessions=average_estimated_possessions,
        points_per_estimated_possessions=points_per_estimated_possessions,
        average_team_turnovers=average_team_turnovers,
        average_total_turnovers=average_total_turnovers,
        two_point_field_goals_made=two_point_field_goals_made,
        two_point_field_goals_attempted=two_point_field_goals_attempted,
        average_two_point_field_goals_made=average_two_point_field_goals_made,
        average_two_point_field_goals_attempted=average_two_point_field_goals_attempted,
        two_point_field_goal_percentage=two_point_field_goal_percentage,
        shooting_efficiency=shooting_efficiency,
        scoring_efficiency=scoring_efficiency,
        average_48_field_goals_made=average_48_field_goals_made,
        average_48_field_goals_attempted=average_48_field_goals_attempted,
        average_48_three_point_field_goals_made=average_48_three_point_field_goals_made,
        average_48_three_point_field_goals_attempted=average_48_three_point_field_goals_attempted,
        average_48_free_throws_made=average_48_free_throws_made,
        average_48_free_throws_attempted=average_48_free_throws_attempted,
        average_48_points=average_48_points,
        average_48_offensive_rebounds=average_48_offensive_rebounds,
        average_48_assists=average_48_assists,
        average_48_turnovers=average_48_turnovers,
        average_rebounds=average_rebounds,
        average_fouls=average_fouls,
        average_flagrant_fouls=average_flagrant_fouls,
        average_technical_fouls=average_technical_fouls,
        average_ejections=average_ejections,
        average_disqualifications=average_disqualifications,
        assist_turnover_ratio=assist_turnover_ratio,
        steal_foul_ratio=steal_foul_ratio,
        block_foul_ratio=block_foul_ratio,
        average_team_rebounds=average_team_rebounds,
        total_technical_fouls=total_technical_fouls,
        team_assist_turnover_ratio=team_assist_turnover_ratio,
        steal_turnover_ratio=steal_turnover_ratio,
        average_48_rebounds=average_48_rebounds,
        average_48_fouls=average_48_fouls,
        average_48_flagrant_fouls=average_48_flagrant_fouls,
        average_48_technical_fouls=average_48_technical_fouls,
        average_48_ejections=average_48_ejections,
        average_48_disqualifications=average_48_disqualifications,
        double_double=double_double,
        triple_double=triple_double,
        field_goals_made=field_goals_made,
        free_throws_made=free_throws_made,
        three_point_percentage=three_point_percentage,
        three_point_field_goals_made=three_point_field_goals_made,
        team_turnovers=team_turnovers,
        total_turnovers=total_turnovers,
        points_in_paint=points_in_paint,
        brick_index=brick_index,
        fast_break_points=fast_break_points,
        average_field_goals_made=average_field_goals_made,
        turnover_points=turnover_points,
        average_defensive_rebounds=average_defensive_rebounds,
        average_blocks=average_blocks,
        average_steals=average_steals,
        average_48_defensive_rebounds=average_48_defensive_rebounds,
        average_48_blocks=average_48_blocks,
        average_48_steals=average_48_steals,
        largest_lead=largest_lead,
        disqualifications=disqualifications,
        flagrant_fouls=flagrant_fouls,
        fouls=fouls,
        ejections=ejections,
        technical_fouls=technical_fouls,
        vorp=vorp,
        average_minutes=average_minutes,
        nba_rating=nba_rating,
        version=VERSION,
        set_one_points=set_one_points,
        set_two_points=set_two_points,
        set_three_points=set_three_points,
        set_four_points=set_four_points,
        set_five_points=set_five_points,
    )

    ffill(team_ffill, identifier, team_model)

    return team_model



"""ATP TennisData league model."""

from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...league import League
from ...tennisdata.tennisdata_league_model import TennisDataLeagueModel


class ATPTennisDataLeagueModel(TennisDataLeagueModel):
    """ATP TennisData implementation of the league model."""

    def __init__(
        self,
        session: ScrapeSession,
        position: int | None = None,
    ) -> None:
        super().__init__(League.ATP, session, position=position)

    @classmethod
    def name(cls) -> str:
        """The name of the league model."""
        return "atp-tennisdata-league-model"



"""Tests for the game model class."""
import datetime
import unittest

from sportsball.data.game_model import GameModel, VERSION
from sportsball.data.league import League


class TestGameModel(unittest.TestCase):

    def setUp(self):
        self._game_model = GameModel(
            dt=datetime.datetime(2010, 1, 1, 10, 10, 0),
            week=None,
            game_number=None,
            venue=None,
            teams=[],
            end_dt=None,
            attendance=None,
            league=str(League.NBA),
            year=None,
            season_type=None,
            postponed=True,
            play_off=None,
            distance=None,
            dividends=[],
            pot=None,
            umpires=[],
            version=VERSION,
            best_of=None,
        )

    def test_postponed(self):
        self.assertTrue(self._game_model.postponed)



"""Sports Reference umpire model."""

# pylint: disable=too-many-locals,too-many-branches
import datetime
import logging
import os
from urllib.parse import urlparse

import pytest_is_running
from bs4 import BeautifulSoup
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...cache import MEMORY
from ..google.address_exception import AddressException
from ..google.google_address_model import create_google_address_model
from ..umpire_model import VERSION, UmpireModel
from .sportsreference_venue_model import create_sportsreference_venue_model


def _create_sportsreference_umpire_model(
    url: str,
    session: ScrapeSession,
    dt: datetime.datetime,
    version: str,
) -> UmpireModel:
    response = session.get(url)
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")

    name = None
    for h1 in soup.find_all("h1"):
        for span in h1.find_all("span"):
            name = span.get_text().strip()
            break
        if name is not None:
            break

    if name is None:
        for h1 in soup.find_all("h1"):
            name = h1.get_text().strip()

    if name is None:
        logging.error(response.text)
        logging.error("error url = %s", url)
        raise ValueError("name is null")

    o = urlparse(url)
    last_component = o.path.split("/")[-1]
    identifier, _ = os.path.splitext(last_component)

    birth_date = None
    for span in soup.find_all("span", {"id": "necro-birth"}):
        birth_date = parse(span.get("data-birth"))

    birth_address = None
    for p in soup.find_all("p"):
        p_text = p.get_text().strip()
        if "Born:" not in p_text:
            continue
        address_text = p_text.split("in ")[-1].strip().split("College:")[0].strip()
        try:
            birth_address = create_google_address_model(
                query=address_text, session=session, dt=None
            )
        except AddressException as exc:
            logging.warning("Failed to find birth address: %s", str(exc))
        break

    high_school = None
    for p in soup.find_all("p"):
        p_text = p.get_text().strip()
        if "High School:" not in p_text:
            continue
        address_text = p_text.split("in ")[-1].strip().split("College:")[0].strip()
        try:
            high_school = create_sportsreference_venue_model(
                venue_name=address_text, session=session, dt=None
            )
        except AddressException as exc:
            logging.warning("Failed to find high school: %s", str(exc))
        break

    return UmpireModel(
        identifier=identifier,
        name=name,
        birth_date=birth_date,
        age=None if birth_date is None else relativedelta(birth_date, dt).years,
        birth_address=birth_address,
        high_school=high_school,
        version=version,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_umpire_model(
    url: str,
    session: ScrapeSession,
    dt: datetime.datetime,
    version: str,
) -> UmpireModel:
    return _create_sportsreference_umpire_model(
        url=url, session=session, dt=dt, version=version
    )


def create_sportsreference_umpire_model(
    url: str,
    session: ScrapeSession,
    dt: datetime.datetime,
) -> UmpireModel:
    """Create a sports reference umpire model."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_umpire_model(
            url=url, session=session, dt=dt, version=VERSION
        )
    with session.cache_disabled():
        return _create_sportsreference_umpire_model(
            url=url, session=session, dt=dt, version=VERSION
        )



"""The prototype class defining how to interface to the league."""

# pylint: disable=line-too-long
import datetime
import logging
import threading
from typing import Any, Callable, Iterator, Type, get_args, get_origin

import pandas as pd
import tqdm
from flatten_json import flatten  # type: ignore
from pydantic import BaseModel
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from .address_model import ADDRESS_TIMEZONE_COLUMN
from .delimiter import DELIMITER
from .field_type import FieldType
from .game_model import GAME_DT_COLUMN, VENUE_COLUMN_PREFIX, GameModel
from .league import League
from .model import Model
from .venue_model import VENUE_ADDRESS_COLUMN

LEAGUE_COLUMN = "league"
SHUTDOWN_FLAG = threading.Event()
TIMEOUT_DT = datetime.datetime.now() + datetime.timedelta(days=365)


def _clear_column_list(df: pd.DataFrame) -> pd.DataFrame:
    def has_list(x):
        return any(isinstance(i, list) for i in x)

    mask = df.apply(has_list)
    cols = df.columns[mask].tolist()
    return df.drop(columns=cols)


def _reduce_memory_usage(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.columns:
        if df[col].dtype == "int64":
            df[col] = pd.to_numeric(df[col], downcast="integer")
        elif df[col].dtype == "float64":
            df[col] = pd.to_numeric(df[col], downcast="float")
    return df


def _normalize_tz(df: pd.DataFrame) -> pd.DataFrame:
    tz_column = DELIMITER.join(
        [VENUE_COLUMN_PREFIX, VENUE_ADDRESS_COLUMN, ADDRESS_TIMEZONE_COLUMN]
    )
    if tz_column not in df.columns.values.tolist():
        return df
    df = df.dropna(subset=tz_column)

    tqdm.tqdm.pandas(desc="Timezone Conversions")

    # Check each row to see if they have the correct timezone
    dt_cols = set()

    def apply_tz(row: pd.Series) -> pd.Series:
        if tz_column not in row:
            return row
        tz = row[tz_column]
        if pd.isnull(tz):
            return row

        datetime_cols = {
            col for col, val in row.items() if isinstance(val, pd.Timestamp)
        }
        datetime_cols.add(GAME_DT_COLUMN)
        for col in datetime_cols:
            dt_cols.add(col)
            dt = row[col]  # type: ignore
            if isinstance(dt, (datetime.date, datetime.datetime)):
                dt = pd.to_datetime(dt)
            if dt.tz is None:
                row[col] = dt.tz_localize(
                    tz, ambiguous=True, nonexistent="shift_forward"
                )
            elif str(dt.tz) != str(tz):
                row[col] = dt.tz_convert(tz)

        return row

    for dt_col in dt_cols:
        df[dt_col] = pd.to_datetime(df[dt_col])

    return df.progress_apply(apply_tz, axis=1)  # type: ignore


def _find_nested_paths(
    checker: Callable[[Any], bool],
    model_class: type[BaseModel],
    cols: list[str],
) -> list[str]:
    def any_column_contains(substr: str) -> bool:
        for col in cols:
            if substr in col:
                return True
        return False

    nested_paths = []
    for field_name, field in model_class.model_fields.items():
        if checker(field):
            nested_paths.append(field_name)
        else:
            if issubclass(
                get_origin(field.annotation) or field.annotation,  # type: ignore
                BaseModel,  # type: ignore
            ):
                nested_paths.extend(
                    [
                        DELIMITER.join([field_name, x])
                        for x in _find_nested_paths(
                            checker,
                            field.annotation,  # type: ignore
                            cols,
                        )  # type: ignore
                    ]
                )
            elif get_origin(field.annotation) == list and issubclass(
                get_args(field.annotation)[0], BaseModel
            ):
                i = 0
                while any_column_contains(DELIMITER.join([field_name, str(i)])):
                    nested_paths.extend(
                        [
                            DELIMITER.join([field_name, str(i), x])
                            for x in _find_nested_paths(
                                checker, get_args(field.annotation)[0], cols
                            )
                        ]
                    )
                    i += 1
    return nested_paths


def _find_nested_field_type_paths(
    field_type: str, model_class: type[BaseModel], cols: list[str]
) -> list[str]:
    def _check_field_type(field: Any) -> bool:
        nested_field_type = (
            field.json_schema_extra.get("type")  # type: ignore
            if field.json_schema_extra
            else None
        )
        return nested_field_type == field_type

    return _find_nested_paths(_check_field_type, model_class, cols)


def _find_nested_field_typing_paths(
    field_typing: Type, model_class: type[BaseModel], cols: list[str]
) -> list[str]:
    def _check_field_typing(field: Any) -> bool:
        return field.annotation == field_typing

    return _find_nested_paths(_check_field_typing, model_class, cols)


def _print_memory_usage(df: pd.DataFrame) -> None:
    mem_usage = df.memory_usage(deep=True, index=False)
    summary = pd.DataFrame({"dtype": df.dtypes, "memory_usage_bytes": mem_usage})
    summary["memory_usage_MB"] = summary["memory_usage_bytes"] / (1024**2)
    summary_sorted = summary.sort_values("memory_usage_bytes", ascending=False)
    logging.info(summary_sorted.head(50))


def needs_shutdown() -> bool:
    """Whether the system needs to shutdown."""
    if SHUTDOWN_FLAG.is_set():
        return True
    if TIMEOUT_DT < datetime.datetime.now():
        return True
    return False


class LeagueModel(Model):
    """The prototype league model class."""

    def __init__(
        self,
        league: League,
        session: ScrapeSession,
        position: int | None = None,
    ) -> None:
        super().__init__(session)
        self._league = league
        self.position = position

    @classmethod
    def name(cls) -> str:
        """The name of the league model."""
        raise NotImplementedError("name is not implemented by parent class.")

    @property
    def games(self) -> Iterator[GameModel]:
        """Find all the games in this league."""
        raise NotImplementedError("games not implemented by LeagueModel parent class.")

    @property
    def league(self) -> League:
        """Return the league this league model represents."""
        return self._league

    def to_frame(self) -> pd.DataFrame:
        """Render the league as a dataframe."""
        jsonl: list[dict[str, Any]] = []
        cols: set[str] = set()
        for game in tqdm.tqdm(self.games, desc="Games"):
            game_dict = flatten(
                game.model_dump(by_alias=True, exclude_none=True, exclude_unset=True),
                DELIMITER,
            )
            jsonl.append(game_dict)
            cols |= set(game_dict.keys())

        data: dict[str, Any] = {x: [] for x in cols}
        for json_dict in jsonl:
            for col in cols:
                data[col].append(json_dict.get(col))

        categorical_cols = set(
            _find_nested_field_type_paths(FieldType.CATEGORICAL, GameModel, list(cols))
        )
        for k in data:
            if k in categorical_cols:
                data[k] = pd.Categorical(data[k])

        datetime_cols = set(
            _find_nested_field_typing_paths(datetime.datetime, GameModel, list(cols))
        )
        datetime_cols |= set(
            _find_nested_field_typing_paths(datetime.date, GameModel, list(cols))
        )
        for k in data:
            if k in datetime_cols:
                data[k] = pd.to_datetime(data[k], utc=True, errors="coerce")

        df = pd.DataFrame(data)

        lookahead_columns = set(
            _find_nested_field_type_paths(
                FieldType.LOOKAHEAD, GameModel, df.columns.values.tolist()
            )
        )
        df.attrs[str(FieldType.LOOKAHEAD)] = list(
            set(df.columns.values) & lookahead_columns
        )
        df.attrs[str(FieldType.ODDS)] = list(
            set(df.columns.values)
            & set(
                _find_nested_field_type_paths(
                    FieldType.ODDS, GameModel, df.columns.values.tolist()
                )
            )
        )
        df.attrs[str(FieldType.POINTS)] = list(
            set(df.columns.values)
            & set(
                _find_nested_field_type_paths(
                    FieldType.POINTS, GameModel, df.columns.values.tolist()
                )
            )
        )
        df.attrs[str(FieldType.TEXT)] = list(
            set(df.columns.values)
            & set(
                _find_nested_field_type_paths(
                    FieldType.TEXT, GameModel, df.columns.values.tolist()
                )
            )
        )
        df.attrs[str(FieldType.CATEGORICAL)] = list(
            set(df.columns.values) & categorical_cols
        )
        df.attrs[str(FieldType.LOOKAHEAD)] = list(
            set(df.attrs[str(FieldType.LOOKAHEAD)])
            | set(df.attrs[str(FieldType.POINTS)])
        )

        for categorical_column in df.attrs[str(FieldType.CATEGORICAL)]:
            df[categorical_column] = df[categorical_column].astype("category")

        df = _normalize_tz(df)

        if GAME_DT_COLUMN in df.columns.values:
            df = df.sort_values(
                by=GAME_DT_COLUMN,
                ascending=True,
            )
        df = _clear_column_list(df)
        df = df.sort_values(by="dt")
        df = df.reset_index()

        df = _reduce_memory_usage(
            df[sorted(df.columns.values.tolist())].dropna(axis=1, how="all")
        )
        # _print_memory_usage(df)
        return df



"""Tennisdata game model."""

# pylint: disable=too-many-arguments,duplicate-code
import datetime

import pytest_is_running
from dateutil.parser import parse
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...cache import MEMORY
from ..game_model import VERSION as GAME_VERSION
from ..game_model import GameModel
from ..league import League
from ..team_model import VERSION as TEAM_VERSION
from ..venue_model import VERSION
from .tennisdata_team_model import create_tennisdata_team_model
from .tennisdata_venue_model import create_tennisdata_venue_model


def _create_tennisdata_game_model(
    location: str,
    dt: datetime.datetime,
    court: str,
    surface: str,
    best_of: str,
    winner: str,
    loser: str,
    winner_rank: str | None,
    loser_rank: str | None,
    winner_total_points: str | None,
    loser_total_points: str | None,
    winner_points_set_1: str | None,
    loser_points_set_1: str | None,
    winner_points_set_2: str | None,
    loser_points_set_2: str | None,
    winner_points_set_3: str | None,
    loser_points_set_3: str | None,
    winner_points_set_4: str | None,
    loser_points_set_4: str | None,
    winner_points_set_5: str | None,
    loser_points_set_5: str | None,
    winner_sets: str | None,
    loser_sets: str | None,
    winner_odds: str | None,
    loser_odds: str | None,
    session: ScrapeSession,
    league: League,
    version: str,
) -> GameModel:
    winner_team_model = create_tennisdata_team_model(
        name=winner,
        points=int(winner_sets) if winner_sets is not None else None,
        rank=int(winner_rank) if winner_rank is not None else None,
        total_points=int(winner_total_points)
        if winner_total_points is not None
        else None,
        set_one_points=int(winner_points_set_1)
        if winner_points_set_1 is not None
        else None,
        set_two_points=int(winner_points_set_2)
        if winner_points_set_2 is not None
        else None,
        set_three_points=int(winner_points_set_3)
        if winner_points_set_3 is not None
        else None,
        set_four_points=int(winner_points_set_4)
        if winner_points_set_4 is not None
        else None,
        set_five_points=int(winner_points_set_5)
        if winner_points_set_5 is not None
        else None,
        odds=winner_odds,
        session=session,
        dt=dt,
        league=league,
        version=TEAM_VERSION,
    )
    loser_team_model = create_tennisdata_team_model(
        name=loser,
        points=int(loser_sets) if loser_sets is not None else None,
        rank=int(loser_rank) if loser_rank is not None else None,
        total_points=int(loser_total_points)
        if loser_total_points is not None
        else None,
        set_one_points=int(loser_points_set_1)
        if loser_points_set_1 is not None
        else None,
        set_two_points=int(loser_points_set_2)
        if loser_points_set_2 is not None
        else None,
        set_three_points=int(loser_points_set_3)
        if loser_points_set_3 is not None
        else None,
        set_four_points=int(loser_points_set_4)
        if loser_points_set_4 is not None
        else None,
        set_five_points=int(loser_points_set_5)
        if loser_points_set_5 is not None
        else None,
        odds=loser_odds,
        session=session,
        dt=dt,
        league=league,
        version=TEAM_VERSION,
    )
    return GameModel(
        dt=dt,
        week=None,
        game_number=None,
        venue=create_tennisdata_venue_model(
            venue=location,
            session=session,
            dt=dt,
            court=court,
            surface=surface,
            version=VERSION,
        ),
        teams=sorted([winner_team_model, loser_team_model], key=lambda x: x.name),
        end_dt=None,
        attendance=None,
        league=str(league),
        year=None,
        season_type=None,
        postponed=None,
        play_off=None,
        distance=None,
        dividends=[],
        pot=None,
        version=version,
        umpires=[],
        best_of=int(best_of),
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_tennisdata_game_model(
    location: str,
    dt: datetime.datetime,
    court: str,
    surface: str,
    best_of: str,
    winner: str,
    loser: str,
    winner_rank: str | None,
    loser_rank: str | None,
    winner_total_points: str | None,
    loser_total_points: str | None,
    winner_points_set_1: str | None,
    loser_points_set_1: str | None,
    winner_points_set_2: str | None,
    loser_points_set_2: str | None,
    winner_points_set_3: str | None,
    loser_points_set_3: str | None,
    winner_points_set_4: str | None,
    loser_points_set_4: str | None,
    winner_points_set_5: str | None,
    loser_points_set_5: str | None,
    winner_sets: str | None,
    loser_sets: str | None,
    winner_odds: str | None,
    loser_odds: str | None,
    session: ScrapeSession,
    league: League,
    version: str,
) -> GameModel:
    return _create_tennisdata_game_model(
        location=location,
        dt=dt,
        court=court,
        surface=surface,
        best_of=best_of,
        winner=winner,
        loser=loser,
        winner_rank=winner_rank,
        loser_rank=loser_rank,
        winner_total_points=winner_total_points,
        loser_total_points=loser_total_points,
        winner_points_set_1=winner_points_set_1,
        loser_points_set_1=loser_points_set_1,
        winner_points_set_2=winner_points_set_2,
        loser_points_set_2=loser_points_set_2,
        winner_points_set_3=winner_points_set_3,
        loser_points_set_3=loser_points_set_3,
        winner_points_set_4=winner_points_set_4,
        loser_points_set_4=loser_points_set_4,
        winner_points_set_5=winner_points_set_5,
        loser_points_set_5=loser_points_set_5,
        winner_sets=winner_sets,
        loser_sets=loser_sets,
        winner_odds=winner_odds,
        loser_odds=loser_odds,
        session=session,
        league=league,
        version=version,
    )


def create_tennisdata_game_model(
    location: str,
    date: str,
    court: str,
    surface: str,
    best_of: str,
    winner: str,
    loser: str,
    winner_rank: str | None,
    loser_rank: str | None,
    winner_total_points: str | None,
    loser_total_points: str | None,
    winner_points_set_1: str | None,
    loser_points_set_1: str | None,
    winner_points_set_2: str | None,
    loser_points_set_2: str | None,
    winner_points_set_3: str | None,
    loser_points_set_3: str | None,
    winner_points_set_4: str | None,
    loser_points_set_4: str | None,
    winner_points_set_5: str | None,
    loser_points_set_5: str | None,
    winner_sets: str | None,
    loser_sets: str | None,
    winner_odds: str | None,
    loser_odds: str | None,
    session: ScrapeSession,
    league: League,
) -> GameModel:
    """Create a game model based off tennisdata."""
    dt = parse(date)
    if not pytest_is_running.is_running() and dt < datetime.datetime.now().replace(
        tzinfo=dt.tzinfo
    ) - datetime.timedelta(days=7):
        return _cached_create_tennisdata_game_model(
            location=location,
            dt=dt,
            court=court,
            surface=surface,
            best_of=best_of,
            winner=winner,
            loser=loser,
            winner_rank=winner_rank,
            loser_rank=loser_rank,
            winner_total_points=winner_total_points,
            loser_total_points=loser_total_points,
            winner_points_set_1=winner_points_set_1,
            loser_points_set_1=loser_points_set_1,
            winner_points_set_2=winner_points_set_2,
            loser_points_set_2=loser_points_set_2,
            winner_points_set_3=winner_points_set_3,
            loser_points_set_3=loser_points_set_3,
            winner_points_set_4=winner_points_set_4,
            loser_points_set_4=loser_points_set_4,
            winner_points_set_5=winner_points_set_5,
            loser_points_set_5=loser_points_set_5,
            winner_sets=winner_sets,
            loser_sets=loser_sets,
            winner_odds=winner_odds,
            loser_odds=loser_odds,
            session=session,
            league=league,
            version=GAME_VERSION,
        )
    with session.cache_disabled():
        return _create_tennisdata_game_model(
            location=location,
            dt=dt,
            court=court,
            surface=surface,
            best_of=best_of,
            winner=winner,
            loser=loser,
            winner_rank=winner_rank,
            loser_rank=loser_rank,
            winner_total_points=winner_total_points,
            loser_total_points=loser_total_points,
            winner_points_set_1=winner_points_set_1,
            loser_points_set_1=loser_points_set_1,
            winner_points_set_2=winner_points_set_2,
            loser_points_set_2=loser_points_set_2,
            winner_points_set_3=winner_points_set_3,
            loser_points_set_3=loser_points_set_3,
            winner_points_set_4=winner_points_set_4,
            loser_points_set_4=loser_points_set_4,
            winner_points_set_5=winner_points_set_5,
            loser_points_set_5=loser_points_set_5,
            winner_sets=winner_sets,
            loser_sets=loser_sets,
            winner_odds=winner_odds,
            loser_odds=loser_odds,
            session=session,
            league=league,
            version=GAME_VERSION,
        )



"""Aussportsbetting game model."""

# pylint: disable=too-many-arguments,duplicate-code
import datetime

import pytest_is_running
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...cache import MEMORY
from ..game_model import VERSION as GAME_VERSION
from ..game_model import GameModel
from ..league import League
from ..team_model import VERSION as TEAM_VERSION
from ..venue_model import VERSION
from .aussportsbetting_team_model import create_aussportsbetting_team_model
from .aussportsbetting_venue_model import create_aussportsbetting_venue_model


def _create_aussportsbetting_game_model(
    dt: datetime.datetime,
    home_team: str,
    away_team: str,
    venue: str | None,
    session: ScrapeSession,
    home_points: float,
    away_points: float,
    home_odds: float,
    away_odds: float,
    league: League,
    play_off: bool,
    version: str,
) -> GameModel:
    venue_model = None
    if venue is not None:
        venue_model = create_aussportsbetting_venue_model(
            venue=venue, session=session, dt=dt, version=VERSION
        )
    home_team_model = create_aussportsbetting_team_model(
        name=home_team,
        points=home_points,
        odds=home_odds,
        session=session,
        dt=dt,
        league=league,
        version=TEAM_VERSION,
    )
    away_team_model = create_aussportsbetting_team_model(
        name=away_team,
        points=away_points,
        odds=away_odds,
        session=session,
        dt=dt,
        league=league,
        version=TEAM_VERSION,
    )
    return GameModel(
        dt=dt,
        week=None,
        game_number=None,
        venue=venue_model,
        teams=[home_team_model, away_team_model],  # pyright: ignore
        end_dt=None,
        attendance=None,
        league=str(league),
        year=None,
        season_type=None,
        postponed=None,
        play_off=play_off,
        distance=None,
        dividends=[],
        pot=None,
        version=version,
        umpires=[],
        best_of=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_aussportsbetting_game_model(
    dt: datetime.datetime,
    home_team: str,
    away_team: str,
    venue: str | None,
    session: ScrapeSession,
    home_points: float,
    away_points: float,
    home_odds: float,
    away_odds: float,
    league: League,
    play_off: bool,
    version: str,
) -> GameModel:
    return _create_aussportsbetting_game_model(
        dt,
        home_team,
        away_team,
        venue,
        session,
        home_points,
        away_points,
        home_odds,
        away_odds,
        league,
        play_off,
        version=version,
    )


def create_aussportsbetting_game_model(
    dt: datetime.datetime,
    home_team: str,
    away_team: str,
    venue: str | None,
    session: ScrapeSession,
    home_points: float,
    away_points: float,
    home_odds: float,
    away_odds: float,
    league: League,
    play_off: bool,
) -> GameModel:
    """Create a game model based off aus sports betting."""
    if not pytest_is_running.is_running() and dt < datetime.datetime.now().replace(
        tzinfo=dt.tzinfo
    ) - datetime.timedelta(days=7):
        return _cached_create_aussportsbetting_game_model(
            dt=dt,
            home_team=home_team,
            away_team=away_team,
            venue=venue,
            session=session,
            home_points=home_points,
            away_points=away_points,
            home_odds=home_odds,
            away_odds=away_odds,
            league=league,
            play_off=play_off,
            version=GAME_VERSION,
        )
    with session.cache_disabled():
        return _create_aussportsbetting_game_model(
            dt=dt,
            home_team=home_team,
            away_team=away_team,
            venue=venue,
            session=session,
            home_points=home_points,
            away_points=away_points,
            home_odds=home_odds,
            away_odds=away_odds,
            league=league,
            play_off=play_off,
            version=GAME_VERSION,
        )



"""EPL combined league model."""

# pylint: disable=line-too-long
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.epl_espn_league_model import EPLESPNLeagueModel
from ..footballdata.epl_footballdata_league_model import \
    EPLFootballDataLeagueModel
from ..oddsportal.epl_oddsportal_league_model import EPLOddsPortalLeagueModel

BOLTON_WANDERERS = "358"
WEST_HAM_UNITED = "371"
DERBY_COUNTY = "374"
SUNDERLAND = "366"
NEWCASTLE_UNITED = "361"
SOUTHAMPTON = "376"
CHARLTON_ATHLETIC = "372"
MANCHESTER_UNITED = "360"
IPSWICH_TOWN = "373"
LIVERPOOL = "364"
TOTTENHAM_HOTSPUR = "367"
LEICESTER_CITY = "375"
MIDDLESBOROUGH = "369"
LEEDS_UNITED = "357"
ASTON_VILLA = "362"
CHELSEA = "363"
FULHAM = "370"
BLACKBURN_ROVERS = "365"
EVERTON = "368"
ARSENAL = "359"
EPL_TEAM_IDENTITY_MAP: dict[str, str] = {
    # ESPN
    "358": BOLTON_WANDERERS,
    "371": WEST_HAM_UNITED,
    "374": DERBY_COUNTY,
    "366": SUNDERLAND,
    "361": NEWCASTLE_UNITED,
    "376": SOUTHAMPTON,
    "372": CHARLTON_ATHLETIC,
    "360": MANCHESTER_UNITED,
    "373": IPSWICH_TOWN,
    "364": LIVERPOOL,
    "367": TOTTENHAM_HOTSPUR,
    "375": LEICESTER_CITY,
    "369": MIDDLESBOROUGH,
    "357": LEEDS_UNITED,
    "362": ASTON_VILLA,
    "363": CHELSEA,
    "370": FULHAM,
    "365": BLACKBURN_ROVERS,
    "368": EVERTON,
    "359": ARSENAL,
}
BOLEYN_GROUND = "304"
STADIUM_OF_LIGHT = "194"
ST_MARYS_STADIUM = "303"
OLD_TRAFFORD = "250"
ANFIELD = "192"
FILBERT_STREET = "191"
ELLAND_ROAD = "190"
STAMFORD_BRIDGE = "249"
EWOOD_PARK = "280"
HIGHBURY = "267"
PORTMAN_ROAD = "257"
WHITE_HART_LANE = "195"
THE_RIVERSIDE_STADIUM = "193"
PRIDE_PARK_STADIUM = "189"
VILLA_PARK = "307"
CRAVEN_COTTAGE = "279"
THE_VALLEY = "188"
ST_JAMES_PARK = "308"
TOUGHSHEET_COMMUNITY_STADIUM = "256"
GOODISON_PARK = "253"
EPL_VENUE_IDENTITY_MAP: dict[str, str] = {
    # ESPN
    "304": BOLEYN_GROUND,
    "194": STADIUM_OF_LIGHT,
    "303": ST_MARYS_STADIUM,
    "250": OLD_TRAFFORD,
    "192": ANFIELD,
    "191": FILBERT_STREET,
    "190": ELLAND_ROAD,
    "249": STAMFORD_BRIDGE,
    "280": EWOOD_PARK,
    "267": HIGHBURY,
    "257": PORTMAN_ROAD,
    "195": WHITE_HART_LANE,
    "193": THE_RIVERSIDE_STADIUM,
    "189": PRIDE_PARK_STADIUM,
    "307": VILLA_PARK,
    "279": CRAVEN_COTTAGE,
    "188": THE_VALLEY,
    "308": ST_JAMES_PARK,
    "256": TOUGHSHEET_COMMUNITY_STADIUM,
    "253": GOODISON_PARK,
}
EPL_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class EPLCombinedLeagueModel(CombinedLeagueModel):
    """NBA combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.EPL,
            [
                EPLESPNLeagueModel(session, position=0),
                EPLOddsPortalLeagueModel(session, position=1),
                EPLFootballDataLeagueModel(session, position=2),
                # EPLSportsDBLeagueModel(session, position=3),
                # EPLSportsReferenceLeagueModel(session, position=4),
            ],
            league_filter,
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return EPL_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return EPL_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return EPL_PLAYER_IDENTITY_MAP

    @classmethod
    def name(cls) -> str:
        return "epl-combined-league-model"



"""Bundesliga combined league model."""

# pylint: disable=line-too-long
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.bundesliga_espn_league_model import BundesligaESPNLeagueModel
from ..footballdata.bundesliga_footballdata_league_model import \
    BundesligaFootballDataLeagueModel
from ..oddsportal.bundesliga_oddsportal_league_model import \
    BundesligaOddsPortalLeagueModel

BUNDESLIGA_TEAM_IDENTITY_MAP: dict[str, str] = {}
BUNDESLIGA_VENUE_IDENTITY_MAP: dict[str, str] = {}
BUNDESLIGA_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class BundesligaCombinedLeagueModel(CombinedLeagueModel):
    """Bundesliga combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.BUNDESLIGA,
            [
                BundesligaESPNLeagueModel(session, position=0),
                BundesligaOddsPortalLeagueModel(session, position=1),
                BundesligaFootballDataLeagueModel(session, position=2),
            ],
            league_filter,
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return BUNDESLIGA_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return BUNDESLIGA_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return BUNDESLIGA_PLAYER_IDENTITY_MAP

    @classmethod
    def name(cls) -> str:
        return "bundesliga-combined-league-model"



"""SportsDB venue model."""

# pylint: disable=duplicate-code
import datetime

import pytest_is_running
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...cache import MEMORY
from ..google.google_address_model import create_google_address_model
from ..venue_model import VERSION, VenueModel


def _create_sportsdb_venue_model(
    session: ScrapeSession,
    venue_id: str,
    dt: datetime.datetime,
    version: str,
) -> VenueModel | None:
    if venue_id == "19533":
        venue_id = "17146"
    if venue_id == "21813":
        venue_id = "23848"
    if venue_id == "21642":
        venue_id = "29570"
    if venue_id == "23652":
        venue_id = "29569"
    if venue_id == "23654":
        venue_id = "15874"
    if venue_id == "23655":
        venue_id = "30856"
    if venue_id == "23656":
        venue_id = "29575"
    if venue_id == "23657":
        venue_id = "23720"
    response = session.get(
        f"https://www.thesportsdb.com/api/v1/json/3/lookupvenue.php?id={venue_id}"
    )
    response.raise_for_status()
    venues = response.json()["venues"]
    if venues is None:
        return None
    venue = venues[0]
    name = venue["strVenue"]

    address = create_google_address_model(
        " - ".join(
            [
                x
                for x in [
                    name,
                    venue["strLocation"],
                    venue["strCountry"],
                ]
                if x is not None and x
            ]
        ),
        session,
        dt,
    )

    return VenueModel(
        identifier=venue_id,
        name=name,
        address=address,  # pyright: ignore
        is_grass=None,
        is_indoor=None,
        is_turf=None,
        is_dirt=None,
        is_hard=None,
        version=version,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsdb_venue_model(
    session: ScrapeSession,
    venue_id: str,
    dt: datetime.datetime,
    version: str,
) -> VenueModel | None:
    return _create_sportsdb_venue_model(
        session=session, venue_id=venue_id, dt=dt, version=version
    )


def create_sportsdb_venue_model(
    session: ScrapeSession, venue_id: str, dt: datetime.datetime
) -> VenueModel | None:
    """Create sports DB venue model."""
    if not pytest_is_running.is_running() and dt < datetime.datetime.now().replace(
        tzinfo=dt.tzinfo
    ) - datetime.timedelta(days=7):
        return _cached_create_sportsdb_venue_model(
            session=session, venue_id=venue_id, dt=dt, version=VERSION
        )
    with session.cache_disabled():
        return _create_sportsdb_venue_model(
            session=session, venue_id=venue_id, dt=dt, version=VERSION
        )



"""The enumeration of the different supported positions."""

# pylint: disable=duplicate-code
from enum import StrEnum


class Position(StrEnum):
    """An enumeration over the different positions."""

    FULL_BACK = "FB"
    HALF_BACK = "HB"
    CENTRE = "C"
    HALF_FORWARD = "HF"
    FULL_FORWARD = "FF"
    FOLLOWERS = "FOL"
    INTERCHANGE = "IC"
    EMERGENCY = "EMG"
    BACK_POCKET_LEFT = "BPL"
    BACK_POCKET_RIGHT = "BPR"
    HALF_BACK_FLANK_LEFT = "HBFL"
    CENTRE_HALF_BACK = "CHB"
    HALF_BACK_FLANK_RIGHT = "HBFR"
    RUCK_ROVER = "RR"
    RUCK = "R"
    HALF_FORWARD_FLANK_LEFT = "HFFL"
    CENTRE_HALF_FORWARD = "CHF"
    HALF_FORWARD_FLANK_RIGHT = "HFFR"
    FRONT_POCKET_LEFT = "FPL"
    FRONT_POCKET_RIGHT = "FPR"
    WING_RIGHT = "WR"
    WING_LEFT = "WL"


_POSITIONS = {str(x): x for x in Position}


def position_from_str(position_str: str) -> Position:
    """Find a position from a string."""
    position = _POSITIONS.get(position_str)
    if position is None:
        if position_str == "INT":
            return Position.INTERCHANGE
        if position_str == "RK":
            return Position.RUCK
        raise ValueError(f"Unrecognised position: {position_str}")
    return position



"""AFLW AFLWStats game model."""

import io
import logging
import urllib.parse
from urllib.parse import urlparse

import pandas as pd
import pytest_is_running
from bs4 import BeautifulSoup
from dateutil.parser import parse
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ....cache import MEMORY
from ...game_model import VERSION, GameModel
from ...league import League
from .aflw_aflwstats_team_model import create_aflw_aflwstats_team_model
from .aflw_aflwstats_venue_model import create_aflw_aflwstats_venue_model


def _create_aflw_aflwstats_game_model(
    session: ScrapeSession,
    url: str,
    version: str,
) -> GameModel:
    o = urlparse(url)
    season_year = o.path.split("/")[-1].split("-")[0]

    response = session.get(url)
    response.raise_for_status()

    soup = BeautifulSoup(response.text, "lxml")

    h4 = soup.find("h4")
    if h4 is None:
        raise ValueError("h4 is null")
    h4_text = h4.get_text()
    dt = None
    week = None
    venue_name = None
    for line in h4_text.split("\n"):
        line = line.strip()
        if not line:
            continue
        if "/" in line and ":" in line:
            dt = parse(line)
        elif "Round" in line:
            week = int(line.replace("Round", "").strip())
        elif "Final" in line:
            pass
        elif venue_name is None:
            venue_name = line

    team_urls = []
    player_urls = {}
    for a in soup.find_all("a"):
        a_url = urllib.parse.urljoin(url, a.get("href"))
        if "/team/" in a_url:
            team_urls.append(a_url)
        elif "/player/" in a_url:
            player_urls[a.get_text().strip()] = a_url

    points = []
    for h2 in soup.find_all("h2"):
        for line in h2.get_text().split("\n"):
            line = line.strip()
            if not line:
                continue
            if "." not in line:
                continue
            points.append(float(line.split()[-1]))

    handle = io.StringIO()
    handle.write(response.text)
    handle.seek(0)
    dfs = pd.read_html(handle)
    players = []
    for df in dfs:
        if "Player" not in df.columns.values.tolist():
            continue
        players.append(df.to_dict())

    if dt is None:
        raise ValueError("dt is null")
    if venue_name is None:
        raise ValueError("venue_name is null")

    try:
        return GameModel(
            dt=dt,
            week=week,
            game_number=None,
            venue=create_aflw_aflwstats_venue_model(
                name=venue_name, session=session, dt=dt
            ),
            teams=[
                create_aflw_aflwstats_team_model(
                    team_url=team_url,
                    players=players[count],  # type: ignore
                    points=float(points[count]),
                    session=session,
                    player_urls=player_urls,
                    dt=dt,
                )
                for count, team_url in enumerate(team_urls)
            ],
            end_dt=None,
            attendance=None,
            league=str(League.AFLW),
            year=int(season_year),
            season_type=None,
            postponed=None,
            play_off=None,
            distance=None,
            dividends=[],
            pot=None,
            version=version,
            umpires=[],
        )
    except ValueError as exc:
        logging.error(str(exc))
        logging.error(url)
        raise exc


@MEMORY.cache(ignore=["session"])
def _cached_create_aflw_aflwstats_game_model(
    session: ScrapeSession,
    url: str,
    version: str,
) -> GameModel:
    return _create_aflw_aflwstats_game_model(
        session=session,
        url=url,
        version=version,
    )


def create_aflw_aflwstats_game_model(
    session: ScrapeSession,
    url: str,
) -> GameModel:
    """Create a game model from AFLWStats."""
    if not pytest_is_running.is_running():
        return _cached_create_aflw_aflwstats_game_model(
            session=session,
            url=url,
            version=VERSION,
        )
    with session.cache_disabled():
        return _create_aflw_aflwstats_game_model(
            session=session,
            url=url,
            version=VERSION,
        )



"""ESPN league model."""

# pylint: disable=too-many-locals,too-many-arguments,line-too-long,too-many-branches,too-many-statements
import datetime
import logging
from typing import Any, Iterator
from urllib.parse import urlparse

import requests
import tqdm
from dateutil.parser import parse
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ..game_model import GameModel
from ..league import League
from ..league_model import SHUTDOWN_FLAG, LeagueModel, needs_shutdown
from ..season_type import SeasonType
from .espn_game_model import create_espn_game_model


def _season_type_from_name(name: str) -> SeasonType:
    if (
        name == "Regular Season"
        or "English Premier League" in name
        or "Group Stage" in name
        or "Spring Regular Season" in name
        or "Round of" in name
        or "Second Round" in name
        or "2015-2016 Barclays Premier League" in name
        or "2014-2015 Barclays Premier League" in name
        or "2013-2014 Barclays Premier League" in name
        or "2012-2013 Barclays Premier League" in name
        or "2011-2012 Barclays Premier League" in name
        or "2010-2011 Barclays Premier League" in name
        or "2009-2010 Barclays Premier League" in name
        or "2008-2009 Barclays Premier League" in name
        or "2007-2008 Barclays Premier League" in name
        or "2006-2007 Barclays Premier League" in name
        or "2005-2006 Barclays Premier League" in name
        or "2004-2005 Barclays Premier League" in name
        or "2003-2004 Barclays Premier League" in name
        or "2002-2003 Barclays Premier League" in name
        or "2001-2002 Barclays Premier League" in name
        or "2025-26 LALIGA" in name
        or "2025-26 German Bundesliga" in name
        or "2024-25 German Bundesliga" in name
    ):
        return SeasonType.REGULAR
    if name == "Preseason" or "Spring Training" in name:
        return SeasonType.PRESEASON
    if (
        name == "Postseason"
        or "Quarterfinals" in name
        or "Semifinals" in name
        or "3rd-Place" in name
        or "Final" in name
        or "Third Place" in name
        or "Quarter-finals" in name
        or "Semi-finals" in name
        or "Play-In Season" in name
        or "Spring Postseason" in name
        or "Qualifying Finals" in name
    ):
        return SeasonType.POSTSEASON
    if name == "Off Season":
        return SeasonType.OFFSEASON
    raise ValueError(f"Unrecognised season name: {name}")


class ESPNLeagueModel(LeagueModel):
    """ESPN implementation of the league model."""

    def __init__(
        self,
        start_url: str,
        league: League,
        session: ScrapeSession,
        position: int | None = None,
    ) -> None:
        super().__init__(league, session, position=position)
        self._start_url = start_url

    @classmethod
    def name(cls) -> str:
        return "espn-league-model"

    @classmethod
    def position_validator(cls) -> dict[str, str]:
        """A dictionary that contains the mapping from positions to standard positions."""
        raise NotImplementedError(
            "position_validator is not implemented by parent class"
        )

    def _produce_game(
        self,
        cache_disabled: bool,
        event_item: dict[str, Any],
        week_count: int | None,
        game_number: int,
        season_type_json: dict[str, Any],
        pbar: tqdm.tqdm,
    ) -> Iterator[GameModel]:
        event = event_item
        if "$ref" in event_item:
            if cache_disabled:
                with self.session.cache_disabled():
                    event_response = self.session.get(event_item["$ref"])
            else:
                event_response = self.session.get(event_item["$ref"])
            event_response.raise_for_status()
            event = event_response.json()

        competitions = event.get("competitions", [])
        if not competitions:
            for grouping in event.get("groupings", []):
                competitions.extend(grouping["competitions"])

        for competition in competitions:
            if "status" in competition:
                status_response = self.session.get(competition["status"]["$ref"])
                status_response.raise_for_status()
                status = status_response.json()
                if not status["type"]["completed"]:
                    continue
            game_model = create_espn_game_model(
                competition,
                week_count,
                game_number,
                self.session,
                self.league,
                season_type_json.get("year"),
                _season_type_from_name(season_type_json["name"]),
                self.position_validator(),
            )
            game_number += 1
            pbar.update(1)
            pbar.set_description(
                f"ESPN {game_model.year} - {game_model.season_type} - {game_model.dt}"
            )
            yield game_model

    def _produce_games(
        self,
        week: dict[str, Any],
        week_count: int,
        season_type_json: dict[str, Any],
        pbar: tqdm.tqdm,
        cache_disabled: bool,
    ) -> Iterator[GameModel]:
        events_page = 1
        events_count = 0
        while True:
            if "events" not in week:
                break
            events_response = self.session.get(
                week["events"]["$ref"] + f"&page={events_page}"
            )
            events_response.raise_for_status()
            events = events_response.json()
            for event_item in events["items"]:
                for game_model in self._produce_game(
                    event_item=event_item,
                    game_number=events_count,
                    week_count=week_count,
                    cache_disabled=cache_disabled,
                    season_type_json=season_type_json,
                    pbar=pbar,
                ):
                    events_count += 1
                    yield game_model
            if events_page >= events["pageCount"]:
                break
            events_page += 1
        qbr_page = 1
        qbr_count = 0
        while True:
            if "qbr" not in week:
                break
            qbr_response = self.session.get(week["qbr"]["$ref"] + f"&page={qbr_page}")
            qbr = qbr_response.json()
            for qbr_item in qbr["items"]:
                for game_model in self._produce_game(
                    event_item=qbr_item["event"],
                    game_number=qbr_count,
                    week_count=week_count,
                    cache_disabled=cache_disabled,
                    season_type_json=season_type_json,
                    pbar=pbar,
                ):
                    yield game_model
                    qbr_count += 1
            if qbr_page >= qbr["pageCount"]:
                break
            qbr_page += 1

    def _produce_week_games(
        self,
        season_type_json: dict[str, Any],
        page: int,
        pbar: tqdm.tqdm,
        cache_disabled: bool,
    ) -> Iterator[GameModel]:
        found_pages = False
        if "weeks" in season_type_json:
            game_page = 1
            week_count = 0
            while True:
                if cache_disabled:
                    with self.session.cache_disabled():
                        weeks_response = self.session.get(
                            season_type_json["weeks"]["$ref"] + f"&page={page}"
                        )
                else:
                    weeks_response = self.session.get(
                        season_type_json["weeks"]["$ref"] + f"&page={page}"
                    )
                weeks_response.raise_for_status()
                weeks = weeks_response.json()
                for item in weeks["items"]:
                    if needs_shutdown():
                        return
                    if cache_disabled:
                        with self.session.cache_disabled():
                            week_response = self.session.get(item["$ref"])
                    else:
                        week_response = self.session.get(item["$ref"])
                    week_response.raise_for_status()
                    week = week_response.json()
                    for game_model in self._produce_games(
                        week, week_count, season_type_json, pbar, cache_disabled
                    ):
                        found_pages = True
                        if game_model.dt.date() >= datetime.datetime.now().date():
                            continue
                        yield game_model
                    week_count += 1
                if game_page >= weeks["pageCount"]:
                    break
                game_page += 1
        if not found_pages and self.league != League.NFL:
            # Lets check the scoreboards via the calendar
            o = urlparse(season_type_json["$ref"])
            path_components = o.path.split("/")
            sport_slug = path_components[3]
            league_slug = path_components[5]
            start_date = parse(season_type_json["startDate"]).date()
            end_date = parse(season_type_json["endDate"]).date()
            calendar_dates = {
                (start_date + datetime.timedelta(days=x))
                for x in range((end_date - start_date).days)
            }
            events_count = 0
            while calendar_dates:
                calendar_date = calendar_dates.pop()
                if calendar_date > datetime.datetime.now().date() - datetime.timedelta(
                    days=2
                ):
                    continue
                dt = calendar_date.strftime("%Y%m%d")
                url = f"https://site.api.espn.com/apis/site/v2/sports/{sport_slug}/{league_slug}/scoreboard?lang=en&region=us&calendartype=whitelist&limit=100&dates={dt}&league={league_slug}"
                scoreboard_response = self.session.get(url)
                scoreboard_response.raise_for_status()
                try:
                    scoreboard = scoreboard_response.json()
                    calendar_list = scoreboard["leagues"][0]["calendar"]
                    if calendar_list:
                        if isinstance(calendar_list[0], str):
                            declared_calendar_dates = {
                                parse(x).date() for x in calendar_list
                            }
                            calendar_dates &= declared_calendar_dates
                    for event in scoreboard["events"]:
                        if not event:
                            continue
                        try:
                            event_id = event["id"]
                        except KeyError as exc:
                            logging.error(str(exc))
                            logging.error(event)
                            raise exc
                        event_response = self.session.get(
                            f"https://sports.core.api.espn.com/v2/sports/{sport_slug}/leagues/{league_slug}/events/{event_id}?lang=en&region=us"
                        )
                        event_response.raise_for_status()
                        for game_model in self._produce_game(
                            event_item=event_response.json(),
                            game_number=events_count,
                            season_type_json=season_type_json,
                            pbar=pbar,
                            cache_disabled=False,
                            week_count=None,
                        ):
                            yield game_model
                            events_count += 1
                except requests.exceptions.JSONDecodeError as exc:
                    logging.error(scoreboard_response.text)
                    logging.error(url)
                    logging.error(str(exc))

    @property
    def games(self) -> Iterator[GameModel]:
        try:
            with self.session.wayback_disabled():
                page = 1
                first = True
                with tqdm.tqdm(position=self.position) as pbar:
                    while True:
                        if page == 1:
                            with self.session.cache_disabled():
                                response = self.session.get(
                                    self._start_url + f"&page={page}"
                                )
                        else:
                            response = self.session.get(
                                self._start_url + f"&page={page}"
                            )
                        response.raise_for_status()
                        seasons = response.json()
                        for item in seasons.get("items", []):
                            season_response = self.session.get(item["$ref"])
                            season_response.raise_for_status()
                            season_json = season_response.json()

                            for season_item in season_json["types"]["items"]:
                                season_type_response = self.session.get(
                                    season_item["$ref"]
                                )
                                season_type_response.raise_for_status()
                                season_type_json = season_type_response.json()

                                yield from self._produce_week_games(
                                    season_type_json, page, pbar, first
                                )
                            first = False

                        if page >= seasons.get("pageCount", 0):
                            break
                        page += 1
        except Exception as exc:
            SHUTDOWN_FLAG.set()
            raise exc



"""NBA NBA.com player model."""

# pylint: disable=duplicate-code
from typing import Any

from ...player_model import VERSION, PlayerModel
from ...sex import Sex
from ...species import Species
from ..position import position_from_str


def create_nba_nbacom_player_model(player_dict: dict[str, Any]) -> PlayerModel:
    """Create a player model from AFL AFL."""
    return PlayerModel(
        identifier=str(player_dict["personId"]),
        jersey=None,
        kicks=None,
        fumbles=None,
        fumbles_lost=None,
        field_goals=None,
        field_goals_attempted=None,
        offensive_rebounds=None,
        assists=None,
        turnovers=None,
        name=player_dict["playerName"],
        marks=None,
        handballs=None,
        disposals=None,
        goals=None,
        behinds=None,
        hit_outs=None,
        tackles=None,
        rebounds=None,
        insides=None,
        clearances=None,
        clangers=None,
        free_kicks_for=None,
        free_kicks_against=None,
        brownlow_votes=None,
        contested_possessions=None,
        uncontested_possessions=None,
        contested_marks=None,
        marks_inside=None,
        one_percenters=None,
        bounces=None,
        goal_assists=None,
        percentage_played=None,
        birth_date=None,
        age=None,
        species=str(Species.HUMAN),
        handicap_weight=None,
        father=None,
        sex=str(Sex.MALE),
        starting_position=str(position_from_str(player_dict["position"])),
        weight=None,
        birth_address=None,
        owner=None,
        seconds_played=None,
        three_point_field_goals=None,
        three_point_field_goals_attempted=None,
        free_throws=None,
        free_throws_attempted=None,
        defensive_rebounds=None,
        steals=None,
        blocks=None,
        personal_fouls=None,
        points=None,
        game_score=None,
        point_differential=None,
        version=VERSION,
        height=None,
        colleges=[],
        headshot=None,
        forced_fumbles=None,
        fumbles_recovered=None,
        fumbles_recovered_yards=None,
        fumbles_touchdowns=None,
        offensive_two_point_returns=None,
        offensive_fumbles_touchdowns=None,
        defensive_fumbles_touchdowns=None,
        average_gain=None,
        completion_percentage=None,
        completions=None,
        espn_quarterback_rating=None,
        interception_percentage=None,
        interceptions=None,
        long_passing=None,
        misc_yards=None,
        net_passing_yards=None,
        net_total_yards=None,
        passing_attempts=None,
        passing_big_plays=None,
        passing_first_downs=None,
        passing_fumbles=None,
        passing_fumbles_lost=None,
        passing_touchdown_percentage=None,
        passing_touchdowns=None,
        passing_yards=None,
        passing_yards_after_catch=None,
        quarterback_rating=None,
        sacks=None,
        passing_yards_at_catch=None,
        sacks_yards_lost=None,
        net_passing_attempts=None,
        total_offensive_plays=None,
        total_points=None,
        total_touchdowns=None,
        total_yards=None,
        total_yards_from_scrimmage=None,
        two_point_pass=None,
        two_point_pass_attempt=None,
        yards_per_completion=None,
        yards_per_pass_attempt=None,
        net_yards_per_pass_attempt=None,
        long_rushing=None,
        rushing_attempts=None,
        rushing_big_plays=None,
        rushing_first_downs=None,
        rushing_fumbles=None,
        rushing_fumbles_lost=None,
        rushing_touchdowns=None,
        rushing_yards=None,
        stuffs=None,
        stuff_yards_lost=None,
        two_point_rush=None,
        two_point_rush_attempts=None,
        yards_per_rush_attempt=None,
        espn_widereceiver=None,
        long_reception=None,
        receiving_big_plays=None,
        receiving_first_downs=None,
        receiving_fumbles=None,
        receiving_fumbles_lost=None,
        receiving_targets=None,
        receiving_touchdowns=None,
        receiving_yards=None,
        receiving_yards_after_catch=None,
        receiving_yards_at_catch=None,
        receptions=None,
        two_point_receptions=None,
        two_point_reception_attempts=None,
        yards_per_reception=None,
        assist_tackles=None,
        average_interception_yards=None,
        average_sack_yards=None,
        average_stuff_yards=None,
        blocked_field_goal_touchdowns=None,
        blocked_punt_touchdowns=None,
        defensive_touchdowns=None,
        hurries=None,
        kicks_blocked=None,
        long_interception=None,
        misc_touchdowns=None,
        passes_batted_down=None,
        passes_defended=None,
        quarterback_hits=None,
        sacks_assisted=None,
        sacks_unassisted=None,
        sacks_yards=None,
        safeties=None,
        solo_tackles=None,
        stuff_yards=None,
        tackles_for_loss=None,
        tackles_yards_lost=None,
        yards_allowed=None,
        points_allowed=None,
        one_point_safeties_made=None,
        missed_field_goal_return_td=None,
        blocked_punt_ez_rec_td=None,
        interception_touchdowns=None,
        interception_yards=None,
        average_kickoff_return_yards=None,
        average_kickoff_yards=None,
        extra_point_attempts=None,
        extra_point_percentage=None,
        extra_point_blocked=None,
        extra_points_blocked_percentage=None,
        extra_points_made=None,
        fair_catches=None,
        fair_catch_percentage=None,
        field_goal_attempts_max_19_yards=None,
        field_goal_attempts_max_29_yards=None,
        field_goal_attempts_max_39_yards=None,
        field_goal_attempts_max_49_yards=None,
        field_goal_attempts_max_59_yards=None,
        field_goal_attempts_max_99_yards=None,
        field_goal_attempts_above_50_yards=None,
        field_goal_attempt_yards=None,
        field_goals_blocked=None,
        field_goals_blocked_percentage=None,
        field_goals_made=None,
        field_goals_made_max_19_yards=None,
        field_goals_made_max_29_yards=None,
        field_goals_made_max_39_yards=None,
        field_goals_made_max_49_yards=None,
        field_goals_made_max_59_yards=None,
        field_goals_made_max_99_yards=None,
        field_goals_made_above_50_yards=None,
        field_goals_made_yards=None,
        field_goals_missed_yards=None,
        kickoff_out_of_bounds=None,
        kickoff_returns=None,
        kickoff_returns_touchdowns=None,
        kickoff_return_yards=None,
        kickoffs=None,
        kickoff_yards=None,
        long_field_goal_attempt=None,
        long_field_goal_made=None,
        long_kickoff=None,
        total_kicking_points=None,
        touchback_percentage=None,
        touchbacks=None,
        defensive_fumble_returns=None,
        defensive_fumble_return_yards=None,
        fumble_recoveries=None,
        fumble_recovery_yards=None,
        kick_return_fair_catches=None,
        kick_return_fair_catch_percentage=None,
        kick_return_fumbles=None,
        kick_return_fumbles_lost=None,
        kick_returns=None,
        kick_return_touchdowns=None,
        kick_return_yards=None,
        long_kick_return=None,
        long_punt_return=None,
        misc_fumble_returns=None,
        misc_fumble_return_yards=None,
        opposition_fumble_recoveries=None,
        opposition_fumble_recovery_yards=None,
        opposition_special_team_fumble_returns=None,
        opposition_special_team_fumble_return_yards=None,
        punt_return_fair_catches=None,
        punt_return_fair_catch_percentage=None,
        punt_return_fumbles=None,
        punt_return_fumbles_lost=None,
        punt_returns=None,
        punt_returns_started_inside_the_10=None,
        punt_returns_started_inside_the_20=None,
        punt_return_touchdowns=None,
        special_team_fumble_returns=None,
        yards_per_kick_return=None,
        yards_per_punt_return=None,
        yards_per_return=None,
        average_punt_return_yards=None,
        gross_average_punt_yards=None,
        long_punt=None,
        net_average_punt_yards=None,
        punts=None,
        punts_blocked=None,
        punts_blocked_percentage=None,
        punts_inside_10=None,
        punts_inside_10_percentage=None,
        punts_inside_20=None,
        punts_inside_20_percentage=None,
        punts_over_50=None,
        punt_yards=None,
        defensive_points=None,
        misc_points=None,
        return_touchdowns=None,
        total_two_point_conversions=None,
        passing_touchdowns_9_yards=None,
        passing_touchdowns_19_yards=None,
        passing_touchdowns_29_yards=None,
        passing_touchdowns_39_yards=None,
        passing_touchdowns_49_yards=None,
        passing_touchdowns_above_50_yards=None,
        receiving_touchdowns_9_yards=None,
        receiving_touchdowns_19_yards=None,
        receiving_touchdowns_29_yards=None,
        receiving_touchdowns_39_yards=None,
        punt_return_yards=None,
        receiving_touchdowns_49_yards=None,
        receiving_touchdowns_above_50_yards=None,
        rushing_touchdowns_9_yards=None,
        rushing_touchdowns_19_yards=None,
        rushing_touchdowns_29_yards=None,
        rushing_touchdowns_39_yards=None,
        rushing_touchdowns_49_yards=None,
        rushing_touchdowns_above_50_yards=None,
        penalties_in_minutes=None,
        even_strength_goals=None,
        power_play_goals=None,
        short_handed_goals=None,
        game_winning_goals=None,
        even_strength_assists=None,
        power_play_assists=None,
        short_handed_assists=None,
        shots_on_goal=None,
        shooting_percentage=None,
        shifts=None,
        time_on_ice=None,
        decision=None,
        goals_against=None,
        shots_against=None,
        saves=None,
        save_percentage=None,
        shutouts=None,
        individual_corsi_for_events=None,
        on_shot_ice_for_events=None,
        on_shot_ice_against_events=None,
        corsi_for_percentage=None,
        relative_corsi_for_percentage=None,
        offensive_zone_starts=None,
        defensive_zone_starts=None,
        offensive_zone_start_percentage=None,
        hits=None,
        true_shooting_percentage=None,
        at_bats=None,
        runs_scored=None,
        runs_batted_in=None,
        bases_on_balls=None,
        strikeouts=None,
        plate_appearances=None,
        hits_at_bats=None,
        obp=None,
        slg=None,
        ops=None,
        pitches=None,
        strikes=None,
        win_probability_added=None,
        average_leverage_index=None,
        wpa_plus=None,
        wpa_minus=None,
        cwpa=None,
        acli=None,
        re24=None,
        putouts=None,
        innings_pitched=None,
        earned_runs=None,
        home_runs=None,
        era=None,
        batters_faced=None,
        strikes_by_contact=None,
        strikes_swinging=None,
        strikes_looking=None,
        ground_balls=None,
        fly_balls=None,
        line_drives=None,
        inherited_runners=None,
        inherited_scores=None,
        effective_field_goal_percentage=None,
        penalty_kicks_made=None,
        penalty_kicks_attempted=None,
        shots_total=None,
        shots_on_target=None,
        yellow_cards=None,
        red_cards=None,
        touches=None,
        expected_goals=None,
        non_penalty_expected_goals=None,
        expected_assisted_goals=None,
        shot_creating_actions=None,
        goal_creating_actions=None,
        passes_completed=None,
        passes_attempted=None,
        pass_completion=None,
        progressive_passes=None,
        carries=None,
        progressive_carries=None,
        take_ons_attempted=None,
        successful_take_ons=None,
        total_passing_distance=None,
        progressive_passing_distance=None,
        passes_completed_short=None,
        passes_attempted_short=None,
        pass_completion_short=None,
        passes_completed_medium=None,
        passes_attempted_medium=None,
        pass_completion_medium=None,
        passes_completed_long=None,
        passes_attempted_long=None,
        pass_completion_long=None,
        expected_assists=None,
        key_passes=None,
        passes_into_final_third=None,
        passes_into_penalty_area=None,
        crosses_into_penalty_area=None,
        live_ball_passes=None,
        dead_ball_passes=None,
        passes_from_free_kicks=None,
        through_balls=None,
        switches=None,
        crosses=None,
        throw_ins_taken=None,
        corner_kicks=None,
        inswinging_corner_kicks=None,
        outswinging_corner_kicks=None,
        straight_corner_kicks=None,
        passes_offside=None,
        passes_blocked=None,
        tackles_won=None,
        tackles_in_defensive_third=None,
        tackles_in_middle_third=None,
        tackles_in_attacking_third=None,
        dribblers_tackled=None,
        dribbles_challenged=None,
        percent_of_dribblers_tackled=None,
        challenges_lost=None,
        shots_blocked=None,
        tackles_plus_interceptions=None,
        errors=None,
        touches_in_defensive_penalty_area=None,
        touches_in_defensive_third=None,
        touches_in_middle_third=None,
        touches_in_attacking_third=None,
        touches_in_attacking_penalty_area=None,
        live_ball_touches=None,
        successful_take_on_percentage=None,
        times_tackled_during_take_ons=None,
        tackled_during_take_on_percentage=None,
        total_carrying_distance=None,
        progressive_carrying_distance=None,
        carries_into_final_third=None,
        carries_into_penalty_area=None,
        miscontrols=None,
        dispossessed=None,
        passes_received=None,
        progressive_passes_received=None,
        second_yellow_card=None,
        fouls_committed=None,
        fouls_drawn=None,
        offsides=None,
        penalty_kicks_won=None,
        penalty_kicks_conceded=None,
        own_goals=None,
        ball_recoveries=None,
        aerials_won=None,
        aerials_lost=None,
        percentage_of_aerials_won=None,
        shots_on_target_against=None,
        post_shot_expected_goals=None,
        passes_attempted_minus_goal_kicks=None,
        throws_attempted=None,
        percentage_of_passes_that_were_launched=None,
        average_pass_length=None,
        goal_kicks_attempted=None,
        percentage_of_goal_kicks_that_were_launched=None,
        average_goal_kick_length=None,
        crosses_faced=None,
        crosses_stopped=None,
        percentage_crosses_stopped=None,
        defensive_actions_outside_penalty_area=None,
        average_distance_of_defensive_actions=None,
        three_point_attempt_rate=None,
        batting_style=None,
        bowling_style=None,
        playing_roles=None,
        runs=None,
        balls=None,
        fours=None,
        sixes=None,
        strikerate=None,
        fall_of_wicket_order=None,
        fall_of_wicket_num=None,
        fall_of_wicket_runs=None,
        fall_of_wicket_balls=None,
        fall_of_wicket_overs=None,
        fall_of_wicket_over_number=None,
        ball_over_actual=None,
        ball_over_unique=None,
        ball_total_runs=None,
        ball_batsman_runs=None,
        overs=None,
        maidens=None,
        conceded=None,
        wickets=None,
        economy=None,
        runs_per_ball=None,
        dots=None,
        wides=None,
        no_balls=None,
        free_throw_attempt_rate=None,
        offensive_rebound_percentage=None,
        defensive_rebound_percentage=None,
        total_rebound_percentage=None,
        assist_percentage=None,
        steal_percentage=None,
        block_percentage=None,
        turnover_percentage=None,
        usage_percentage=None,
        offensive_rating=None,
        defensive_rating=None,
        box_plus_minus=None,
        ace_percentage=None,
        double_fault_percentage=None,
        first_serves_in=None,
        first_serve_percentage=None,
        second_serve_percentage=None,
        break_points_saved=None,
        return_points_won_percentage=None,
        winners=None,
        winners_fronthand=None,
        winners_backhand=None,
        unforced_errors=None,
        unforced_errors_fronthand=None,
        unforced_errors_backhand=None,
        serve_points=None,
        serves_won=None,
        serves_aces=None,
        serves_unreturned=None,
        serves_forced_error_percentage=None,
        serves_won_in_three_shots_or_less=None,
        serves_wide_percentage=None,
        serves_body_percentage=None,
        serves_t_percentage=None,
        serves_wide_deuce_percentage=None,
        serves_body_deuce_percentage=None,
        serves_t_deuce_percentage=None,
        serves_wide_ad_percentage=None,
        serves_body_ad_percentage=None,
        serves_t_ad_percentage=None,
        serves_net_percentage=None,
        serves_wide_direction_percentage=None,
        shots_deep_percentage=None,
        shots_deep_wide_percentage=None,
        shots_foot_errors_percentage=None,
        shots_unknown_percentage=None,
        points_won_percentage=None,
        centre_clearances=None,
        disposal_efficiency=None,
        tackles_inside_50=None,
        total_possessions=None,
        score_involvements=None,
        goal_accuracy=None,
        stoppage_clearances=None,
        uncontested_marks=None,
        accurate_crosses=None,
        accurate_long_balls=None,
        accurate_passes=None,
        accurate_through_balls=None,
        cross_percentage=None,
        free_kick_goals=None,
        free_kick_percentage=None,
        free_kick_shots=None,
        game_winning_assists=None,
        headed_goals=None,
        inaccurate_crosses=None,
        inaccurate_long_balls=None,
        inaccurate_passes=None,
        inaccurate_through_balls=None,
        left_footed_shots=None,
        long_ball_percentage=None,
        penalty_kick_goals=None,
        penalty_kick_percentage=None,
        penalty_kicks_missed=None,
        possession_percentage=None,
        possession_time=None,
        right_footed_shots=None,
        shoot_out_goals=None,
        shoot_out_misses=None,
        shoot_out_percentage=None,
        shot_assists=None,
        shot_percentage=None,
        shots_headed=None,
        shots_off_target=None,
        shots_on_post=None,
        through_ball_percentage=None,
        long_balls=None,
        total_passes=None,
        average_rating_from_editor=None,
        average_rating_from_user=None,
        did_not_play=None,
        draws=None,
        goal_difference=None,
        losses=None,
        lost_corners=None,
        minutes=None,
        pass_percentage=None,
        starts=None,
        sub_ins=None,
        sub_outs=None,
        suspensions=None,
        time_ended=None,
        time_started=None,
        win_percentage=None,
        wins=None,
        won_corners=None,
        clean_sheet=None,
        crosses_caught=None,
        goals_conceded=None,
        partial_clean_sheet=None,
        penalty_kick_conceded=None,
        penalty_kick_save_percentage=None,
        penalty_kicks_faced=None,
        penalty_kicks_saved=None,
        punches=None,
        shoot_out_kicks_faced=None,
        shoot_out_kicks_saved=None,
        shoot_out_save_percentage=None,
        shots_faced=None,
        smothers=None,
        unclaimed_crosses=None,
        effective_clearances=None,
        effective_tackles=None,
        ineffective_tackles=None,
        tackle_percentage=None,
        appearances=None,
        average_rating_from_correspondent=None,
        average_rating_from_data_feed=None,
        games_played=None,
        team_games_played=None,
        hit_by_pitch=None,
        rbis=None,
        sac_hits=None,
        stolen_bases=None,
        walks=None,
        catcher_interference=None,
        gidps=None,
        sac_flies=None,
        grand_slam_home_runs=None,
        runners_left_on_base=None,
        triples=None,
        game_winning_rbis=None,
        intentional_walks=None,
        doubles=None,
        caught_stealing=None,
        games_started=None,
        pinch_at_bats=None,
        pinch_hits=None,
        player_rating=None,
        is_qualified=None,
        is_qualified_steals=None,
        total_bases=None,
        projected_home_runs=None,
        extra_base_hits=None,
        runs_created=None,
        batting_average=None,
        pinch_average=None,
        slug_average=None,
        secondary_average=None,
        on_base_percentage=None,
        ground_to_fly_ratio=None,
        runs_created_per_27_outs=None,
        batter_rating=None,
        at_bats_per_home_run=None,
        stolen_base_percentage=None,
        pitches_per_plate_appearance=None,
        isolated_power=None,
        walk_to_strikeout_ratio=None,
        walks_per_plate_appearance=None,
        secondary_average_minus_batting_average=None,
        runs_produced=None,
        runs_ratio=None,
        patience_ratio=None,
        balls_in_play_average=None,
        mlb_rating=None,
        offensive_wins_above_replacement=None,
        wins_above_replacement=None,
        batters_hit=None,
        sacrifice_bunts=None,
        save_opportunities=None,
        finishes=None,
        balks=None,
        holds=None,
        complete_games=None,
        perfect_games=None,
        wild_pitches=None,
        third_innings=None,
        team_earned_runs=None,
        pickoff_attempts=None,
        run_support=None,
        pitches_as_starter=None,
        average_game_score=None,
        quality_starts=None,
        inherited_runners_scored=None,
        opponent_total_bases=None,
        is_qualified_saves=None,
        full_innings=None,
        part_innings=None,
        blown_saves=None,
        innings=None,
        whip=None,
        caught_stealing_percentage=None,
        pitches_per_start=None,
        pitches_per_inning=None,
        run_support_average=None,
        opponent_average=None,
        opponent_slug_average=None,
        opponent_on_base_percentage=None,
        opponent_ops=None,
        strikeouts_per_nine_innings=None,
        strikeout_to_walk_ratio=None,
        tough_losses=None,
        cheap_wins=None,
        save_opportunities_per_win=None,
        pitch_count=None,
        strike_pitch_ratio=None,
        double_plays=None,
        opportunities=None,
        passed_balls=None,
        outfield_assists=None,
        pickoffs=None,
        outs_on_field=None,
        triple_plays=None,
        balls_in_zone=None,
        extra_bases=None,
        outs_made=None,
        catcher_third_innings_played=None,
        catcher_caught_stealing=None,
        catcher_stolen_bases_allowed=None,
        catcher_earned_runs=None,
        is_qualified_catcher=None,
        is_qualified_pitcher=None,
        successful_chances=None,
        total_chances=None,
        full_innings_played=None,
        part_innings_played=None,
        fielding_percentage=None,
        range_factor=None,
        zone_rating=None,
        catcher_caught_stealing_percentage=None,
        catcher_era=None,
        def_warbr=None,
        average_three_point_field_goals_attempted=None,
        average_free_throws_made=None,
        average_free_throws_attempted=None,
        average_points=None,
        average_offensive_rebounds=None,
        average_assists=None,
        average_turnovers=None,
        estimated_possessions=None,
        average_estimated_possessions=None,
        points_per_estimated_possessions=None,
        average_team_turnovers=None,
        average_total_turnovers=None,
        three_point_field_goal_percentage=None,
        two_point_field_goals_made=None,
        two_point_field_goals_attempted=None,
        average_two_point_field_goals_made=None,
        average_two_point_field_goals_attempted=None,
        two_point_field_goal_percentage=None,
        shooting_efficiency=None,
        scoring_efficiency=None,
        average_48_field_goals_made=None,
        average_48_field_goals_attempted=None,
        average_48_three_point_field_goals_made=None,
        average_48_three_point_field_goals_attempted=None,
        average_48_free_throws_made=None,
        average_48_free_throws_attempted=None,
        average_48_points=None,
        average_48_offensive_rebounds=None,
        average_48_assists=None,
        average_48_turnovers=None,
        p40=None,
        a40=None,
        average_rebounds=None,
        average_fouls=None,
        average_flagrant_fouls=None,
        average_technical_fouls=None,
        average_ejections=None,
        average_disqualifications=None,
        assist_turnover_ratio=None,
        steal_foul_ratio=None,
        block_foul_ratio=None,
        average_team_rebounds=None,
        total_technical_fouls=None,
        team_assist_turnover_ratio=None,
        steal_turnover_ratio=None,
        average_48_rebounds=None,
        average_48_fouls=None,
        average_48_flagrant_fouls=None,
        average_48_technical_fouls=None,
        average_48_ejections=None,
        average_48_disqualifications=None,
        r40=None,
        double_double=None,
        triple_double=None,
        free_throws_made=None,
        three_point_percentage=None,
        three_point_field_goals_made=None,
        total_turnovers=None,
        points_in_paint=None,
        brick_index=None,
        average_field_goals_made=None,
        average_field_goals_attempted=None,
        average_three_point_field_goals_made=None,
        average_defensive_rebounds=None,
        average_blocks=None,
        average_steals=None,
        average_48_defensive_rebounds=None,
        average_48_blocks=None,
        average_48_steals=None,
        largest_lead=None,
        disqualifications=None,
        flagrant_fouls=None,
        fouls=None,
        ejections=None,
        technical_fouls=None,
        average_minutes=None,
        nba_rating=None,
        plus_minus=None,
        even_strength_time_on_ice=None,
        shifts_per_game=None,
        production=None,
        shot_differential=None,
        goal_differential=None,
        pim_differential=None,
        rating=None,
        average_goals=None,
        ytd_goals=None,
        shots_in_first_period=None,
        shots_in_second_period=None,
        shots_in_third_period=None,
        shots_overtime=None,
        average_shots=None,
        points_per_game=None,
        power_play_opportunities=None,
        power_play_percentage=None,
        shootout_attempts=None,
        shootout_shot_percentage=None,
        empty_net_goals_for=None,
        shutouts_against=None,
        total_face_offs=None,
        average_goals_against=None,
        average_shots_against=None,
        penalty_kill_percentage=None,
        power_play_goals_against=None,
        short_handed_goals_against=None,
        shootout_saves=None,
        shootout_shots_against=None,
        times_short_handed=None,
        empty_net_goals_against=None,
        overtime_losses=None,
        shots_missed=None,
        minor_penalties=None,
        match_penalties=None,
        misconducts=None,
        game_misconducts=None,
        boarding_penalties=None,
        unsportsmanlike_penalties=None,
        fighting_penalties=None,
        average_fights=None,
        time_between_fights=None,
        instigator_penalties=None,
        charging_penalties=None,
        hooking_penalties=None,
        tripping_penalties=None,
        roughing_penalties=None,
        holding_penalties=None,
        interference_penalties=None,
        slashing_penalties=None,
        high_sticking_penalties=None,
        cross_checking_penalties=None,
        stick_holding_penalties=None,
        goalie_interference_penalties=None,
        elbowing_penalties=None,
        diving_penalties=None,
        takeaways=None,
        even_strength_saves=None,
        power_play_saves=None,
        short_handed_saves=None,
        games=None,
        game_started=None,
        ties=None,
        time_on_ice_per_game=None,
        power_play_time_on_ice=None,
        short_handed_time_on_ice=None,
        faceoffs_won=None,
        faceoffs_lost=None,
        faceoff_percentage=None,
        unassisted_goals=None,
        game_tying_goals=None,
        giveaways=None,
        penalties=None,
        penalty_minutes=None,
        penalty_minutes_against=None,
        major_penalties=None,
        net_passing_yards_per_game=None,
        net_yards_per_game=None,
        passing_yards_per_game=None,
        total_points_per_game=None,
        yards_from_scrimmage_per_game=None,
        yards_per_game=None,
        espn_rb_rating=None,
        rushing_yards_per_game=None,
        receiving_yards_per_game=None,
        two_point_returns=None,
        field_goal_attempts=None,
        special_team_fumble_return_yards=None,
        kick_extra_points=None,
        kick_extra_points_made=None,
        attempts_in_box=None,
        second_assists=None,
        qbr=None,
        attempts_out_box=None,
        adjusted_qbr=None,
        turnover_points=None,
        fantasy_rating=None,
        team_turnovers=None,
        second_chance_points=None,
        fast_break_points=None,
        team_rebounds=None,
        gained=None,
    )



"""EPL combined league model."""

# pylint: disable=line-too-long
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.laliga_espn_league_model import LaLigaESPNLeagueModel
from ..footballdata.laliga_footballdata_league_model import \
    LaLigaFootballDataLeagueModel
from ..oddsportal.laliga_oddsportal_league_model import \
    LaLigaOddsPortalLeagueModel

LALIGA_TEAM_IDENTITY_MAP: dict[str, str] = {}
LALIGA_VENUE_IDENTITY_MAP: dict[str, str] = {}
LALIGA_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class LaLigaCombinedLeagueModel(CombinedLeagueModel):
    """LaLiga combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.LALIGA,
            [
                LaLigaESPNLeagueModel(session, position=0),
                LaLigaOddsPortalLeagueModel(session, position=1),
                LaLigaFootballDataLeagueModel(session, position=2),
            ],
            league_filter,
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return LALIGA_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return LALIGA_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return LALIGA_PLAYER_IDENTITY_MAP

    @classmethod
    def name(cls) -> str:
        return "laliga-combined-league-model"



"""WNBA WNBA.com league model."""

import urllib.parse
from typing import Iterator

from bs4 import BeautifulSoup
from playwright.sync_api import sync_playwright
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ....playwright import ensure_install
from ...game_model import VERSION, GameModel
from ...league import League
from ...league_model import SHUTDOWN_FLAG, LeagueModel
from .wnba_wnbacom_game_model import create_wnba_wnbacom_game_model


class WNBAWNBAComLeagueModel(LeagueModel):
    """WNBA WNBA.com implementation of the league model."""

    def __init__(self, session: ScrapeSession, position: int | None = None) -> None:
        super().__init__(League.WNBA, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "wnba-nflcom-league-model"

    @property
    def games(self) -> Iterator[GameModel]:
        try:
            ensure_install()
            with sync_playwright() as p:
                browser = p.chromium.launch()
                context = browser.new_context()
                page = context.new_page()
                url = "https://www.wnba.com/schedule?season=2025&month=all"
                page.goto(url, wait_until="load")
                soup = BeautifulSoup(page.content(), "lxml")
                game_urls = []
                for a in soup.find_all(
                    "a",
                ):
                    game_url = urllib.parse.urljoin(url, a.get("href"))
                    if not game_url.startswith("https://www.wnba.com/game/"):
                        continue
                    game_urls.append(game_url)
                for game_url in game_urls:
                    yield create_wnba_wnbacom_game_model(
                        url=game_url,
                        session=self.session,
                        playwright=p,
                        version=VERSION,
                    )
        except Exception as exc:
            SHUTDOWN_FLAG.set()
            raise exc



"""The enumeration of the different supported positions."""

from enum import StrEnum


class Position(StrEnum):
    """An enumeration over the different positions."""

    GOALKEEPER = "G"
    CENTRE_DEFENDER_LEFT = "CD-L"
    CENTRE_DEFENDER_RIGHT = "CD-R"
    LEFT_BACK = "LB"
    RIGHT_BACK = "RB"
    ATTACKING_MIDFIELDER = "AM"
    CENTRAL_MIDFIELDER = "CM"
    LEFT_MIDFIELDER = "LM"
    RIGHT_MIDFIELDER = "RM"
    CENTRAL_MIDFIELDER_LEFT = "CM-L"
    FORWARD = "F"
    LEFT_FORWARD = "LF"
    CENTRE_DEFENDER = "CD"
    ATTACKING_MIDFIELDER_LEFT = "AM-L"
    RIGHT_FORWARD = "RF"
    ATTACKING_MIDFIELDER_RIGHT = "AM-R"
    SUBSTITUTE = "SUB"
    CENTRE_MIDFIELDER_RIGHT = "CM-R"
    DEFENSIVE_MIDFIELDER = "DM"
    CENTRE_FORWARD_LEFT = "CF-L"
    CENTRE_FORWARD_RIGHT = "CF-R"
    SWEEPER = "SW"
    MIDFIELDER = "M"
    RIGHT_CENTRAL_MIDFIELDER = "RCM"
    DEFENDER = "D"


_POSITIONS = {str(x): x for x in Position}


def position_from_str(position_str: str) -> Position:
    """Find a position from a string."""
    position = _POSITIONS.get(position_str)
    if position is None:
        raise ValueError(f"Unrecognised position: {position_str}")
    return position



"""Tests for the aussportsbetting game model class."""
import datetime
import os
import unittest

import requests_mock
import requests_cache
from sportsball.data.aussportsbetting.aussportsbetting_game_model import create_aussportsbetting_game_model
from sportsball.data.league import League


class TestAusSportsBettingGameModel(unittest.TestCase):

    def setUp(self):
        self._session = requests_cache.CachedSession(backend="memory")
        self.dir = os.path.dirname(__file__)

    def test_dt(self):
        dt = datetime.datetime(2023, 9, 15, 0, 15)
        with requests_mock.Mocker() as m:
            game_model = create_aussportsbetting_game_model(
                dt,
                "Hello",
                "There",
                None,
                self._session,
                10.0,
                10.0,
                1.0,
                1.0,
                League.NBA,
                True,
            )
            self.assertEqual(game_model.dt, dt)



"""Combined venue model."""

import requests

from ..venue_model import VERSION, VenueModel
from ..wikipedia.wikipedia_venue_model import create_wikipedia_venue_model
from .combined_address_model import create_combined_address_model
from .most_interesting import more_interesting
from .null_check import is_null


def create_combined_venue_model(
    venue_models: list[VenueModel],
    identifier: str | None,
    session: requests.Session,
) -> VenueModel | None:
    """Create a venue model by combining many venue models."""
    if not venue_models or identifier is None:
        return None
    address_model = None
    for venue_model in venue_models:
        address_model = venue_model.address
        if address_model is not None:
            latitude = address_model.latitude
            longitude = address_model.longitude
            if latitude is not None and longitude is not None:
                wikipedia_venue_model = create_wikipedia_venue_model(
                    session,
                    latitude=latitude,
                    longitude=longitude,
                    version=VERSION,
                )
                if wikipedia_venue_model is not None:
                    venue_models.append(wikipedia_venue_model)
                break

    address_models = []
    is_grass = None
    is_indoor = None
    is_turf = None
    is_dirt = None
    is_hard = None
    for venue_model in venue_models:
        venue_model_address = venue_model.address
        if not is_null(venue_model_address):
            address_models.append(venue_model_address)
        is_grass = more_interesting(is_grass, venue_model.is_grass)
        is_indoor = more_interesting(is_indoor, venue_model.is_indoor)
        is_turf = more_interesting(is_turf, venue_model.is_turf)
        is_dirt = more_interesting(is_dirt, venue_model.is_dirt)
        is_hard = more_interesting(is_hard, venue_model.is_hard)
    return VenueModel.model_construct(
        identifier=identifier,
        name=venue_models[0].name,
        address=create_combined_address_model(address_models),  # type: ignore
        is_grass=is_grass,
        is_indoor=is_indoor,
        is_turf=is_turf,
        is_dirt=is_dirt,
        is_hard=is_hard,
        version=VERSION,
    )



"""The bundesliga data sportsball module."""

from .combined.bundesliga_combined_league_model import \
    BundesligaCombinedLeagueModel as BundesligaLeagueModel

__all__ = ("BundesligaLeagueModel",)



"""Tennisdata odds model."""

from ..bet import Bet
from ..team_model import OddsModel
from .tennisdata_bookie_model import create_tennisdata_bookie_model


def create_tennisdata_odds_model(odds: str) -> OddsModel:
    """Create an odds model based off tennisdata."""
    bookie = create_tennisdata_bookie_model()
    return OddsModel(
        odds=float(odds), bookie=bookie, dt=None, canonical=True, bet=str(Bet.WIN)
    )



"""Fallback weather model."""

import datetime

import requests_cache

from ...cache import MEMORY
from ..weather_model import WeatherModel
from .gribstream.gribstream_weather_model import \
    create_gribstream_weather_model
from .openmeteo.openmeteo_weather_model import create_openmeteo_weather_model


@MEMORY.cache(ignore=["session"])
def create_mutli_weather_model(
    session: requests_cache.CachedSession,
    latitude: float,
    longitude: float,
    dt: datetime.datetime,
    tz: str,
) -> WeatherModel | None:
    """Create a weather model by falling back on different providers."""
    weather_model = create_openmeteo_weather_model(session, latitude, longitude, dt, tz)
    if weather_model is None:
        weather_model = create_gribstream_weather_model(
            session, latitude, longitude, dt
        )
    return weather_model
